""" Module to handle composition manipulations and analyses.

"""

import numpy as np
import pandas as pd
import json
from sklearn.preprocessing import MultiLabelBinarizer
from pymatgen.core.periodic_table import Element
from collections import Counter
from itertools import combinations

from scipy.sparse import coo_matrix

import warnings

try:
    from pymatgen import Structure
except ImportError:
    from pymatgen.core.structure import Structure
from pymatgen.analysis.local_env import CrystalNN

from . import extendRDF

def element_indice():
    '''
    Generate 
    '''
    global modified_pettifor, pettifor, periodic_table, periodic_table_78

    # 78 elements without nobal gas and rare earch Ac
    # Put the elements with similar propeties together to give a better representation, see
    # Pettifor, D. G. (1990). "Structure maps in alloy design." 
    # Journal of the Chemical Society, Faraday Transactions 86(8): 1209.
    pettifor = [
        'Cs', 'Rb', 'K', 'Na', 'Li', 
        'Ba', 'Sr', 'Ca', 
        'Yb', 'Eu', 'Y',  'Sc', 'Lu', 'Tm', 'Er', 'Ho', 
        'Dy', 'Tb', 'Gd', 'Sm', 'Pm', 'Nd', 'Pr', 'Ce', 'La', 
        'Zr', 'Hf', 'Ti', 'Nb', 'Ta', 'V',  'Mo', 'W',  'Cr', 'Tc', 'Re', 
        'Mn', 'Fe', 'Os', 'Ru', 'Co', 'Ir', 'Rh', 'Ni', 'Pt', 'Pd', 'Au', 'Ag', 'Cu', 
        'Mg', 'Hg', 'Cd', 'Zn', 'Be', 'Tl', 'In', 'Al', 'Ga', 'Pb', 'Sn', 'Ge', 'Si', 'B',  
        'Bi', 'Sb', 'As', 'P',  'Te', 'Se', 'S', 'C', 'I', 'Br', 'Cl', 'N', 'O', 'F', 'H'
    ]

    # 103 elements, A modified version of Pettifor, see
    # Glawe, H., et al. (2016). New Journal of Physics 18(9): 093011.
    # "The optimal one dimensional periodic table: a modified Pettifor chemical scale from data mining." 
    modified_pettifor = [
        'He', 'Ne', 'Ar', 'Kr', 'Xe', 'Rn', 
        'Fr', 'Cs', 'Rb', 'K', 'Na', 'Li', 'Ra', 'Ba', 'Sr', 'Ca', 
        'Eu', 'Yb', 'Lu', 'Tm', 'Y', 'Er', 'Ho', 'Dy', 'Tb', 'Gd', 'Sm', 'Pm', 'Nd', 'Pr', 'Ce', 'La', 
        'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 
        'Sc', 'Zr', 'Hf', 'Ti', 'Ta', 'Nb', 'V', 'Cr', 'Mo', 'W', 'Re', 
        'Tc', 'Os', 'Ru', 'Ir', 'Rh', 'Pt', 'Pd', 'Au', 'Ag', 'Cu', 
        'Ni', 'Co', 'Fe', 'Mn', 'Mg', 'Zn', 'Cd', 'Hg', 
        'Be', 'Al', 'Ga', 'In', 'Tl', 'Pb', 'Sn', 'Ge', 'Si', 'B', 'C', 
        'N', 'P', 'As', 'Sb', 'Bi', 'Po', 'Te', 'Se', 'S', 'O', 'At', 'I', 'Br', 'Cl', 'F', 'H'
    ]

    # 89 elements in the order of atomic Z number
    # This is the periodic table which DFT pesudopotentials are avaiable
    periodic_table = [
        'H',  'He', 
        'Li', 'Be', 'B',  'C',  'N',  'O',  'F',  'Ne', 
        'Na', 'Mg', 'Al', 'Si', 'P',  'S',  'Cl', 'Ar', 
        'K',  'Ca', 'Sc', 'Ti', 'V',  'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 
        'Rb', 'Sr', 'Y',  'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I',  'Xe', 
        'Cs', 'Ba', 
                    'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 
                          'Hf', 'Ta', 'W',  'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 
    ]

    # 78 elements version by removing nobal gas and six rare-earth Ac-row elements
    periodic_table_78 = [
        'H',  
        'Li', 'Be', 'B',  'C',  'N',  'O',  'F',  
        'Na', 'Mg', 'Al', 'Si', 'P',  'S',  'Cl',  
        'K',  'Ca', 'Sc', 'Ti', 'V',  'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 
        'Rb', 'Sr', 'Y',  'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I',  
        'Cs', 'Ba', 
                    'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 
                          'Hf', 'Ta', 'W',  'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 
    ]

# Always run element_indice to generate ordered species lists
element_indice()

def composition_one_hot(data, method='percentage', index='z_number_78', only_elem_present=False):
    '''
    Make the composition to fix size vector like one hot array
    Note that this function is ONLY suitable for formula with integer number
    of each atom, i.e. a fomula with fractional occuption is not supported

    !!!REINDEX FOR ONLY_TYPE=FASLE NOT IMPLEMENT!!!

    Args:
        data: json dataset of structures from Materials Project
        Method: how the number of atoms is encoded
            percentage: (default) the number of atoms given as percentage, 
                i.e. the sum of the whole formula is 1
            formula: numbers in formula
            only_type: the value of each atom is set to 1
        index: see function element_indice for details
            z_number_78: (default) in the order of atomic number, this is default 
                because the similarity matrix is in this order
            z_number: see periodic_table in function element_indice
            pettifor: see function element_indice for details
            modified_pettifor: see element_indice
            elem_present: the vector only contain the elements presented in the dataset
        only_element_present: Used when the vector is reindexed. If true, only keep the 
            elements presented in the dataset
    Return:
        a pandas dataframe, index is mp-ids, and columns is element symbols
        elem_symbol is a list of elements present in the dataset, in Z number order
    '''
    # define the indice by call element_indice function
    element_indice()

    pero_tab_nums = []
    mp_index = []
    for d in data:
        if 'cif' in d:
            struct = Structure.from_str(d['cif'], fmt='cif')
        elif 'structure' in d:
            struct = d['structure']
        else:
            raise KeyError('No known PyMatGen input in data')
        # the struct.species method give a list of each site, 
        # e.g. for SrTiO3 the output is 
        # [Element Sr, Element Ti, Element O, Element O, Element O]
        pero_tab_nums.append([ x.number for x in struct.species ])
        mp_index.append(d['task_id'])
    
    # use the counter method to one-hot the element numbers
    # considering the output of species method above
    # the data_np is typically using " method == 'formula' "
    elem_vectors = pd.DataFrame([Counter(x) for x in pero_tab_nums])
    elem_vectors = elem_vectors.fillna(0).sort_index(axis=1)

    if method == 'percentage':
        # divide by total number of atoms to the sum will be 1
        # sum all columns in each row, divide row-wise
        elem_vectors = elem_vectors.div(elem_vectors.sum(axis=1), axis=0)
    elif method == 'only_type':
        # set all the non-zero values to 1
        elem_vectors[elem_vectors != 0] = 1

    # get the order of elements in the list and element symbols
    elem_numbers = elem_vectors.columns.tolist()
    # this gives all element symbols present in the dataset
    elem_symbols = [ Element.from_Z(x).name for x in elem_numbers ]
    
    elem_vectors.columns = elem_symbols
    elem_vectors.index = mp_index

    # Now the vectors in data_np is in the order of Z number but only
    # with the elements presented in the dataset
    # we may want to reindex the data_np
    # Note the index here is accutely column names in pandas, not pandas index
    if index != 'elem_present': 
        if index == 'z_number_78':
            elem_vectors = elem_vectors.reindex(columns=periodic_table_78)
        elif index == 'z_number':
            elem_vectors = elem_vectors.reindex(columns=periodic_table)
        elif index == 'pettifor':
            elem_vectors = elem_vectors.reindex(columns=pettifor)
        elif index == 'modified_pettifor':
            elem_vectors = elem_vectors.reindex(columns=modified_pettifor)
            
        if only_elem_present:
            elem_vectors = elem_vectors.dropna()
        else:
            elem_vectors = elem_vectors.fillna(0)

    return elem_vectors, elem_symbols


def elements_count(data):
    '''
    Count the elements distribution histogram in the compounds dataset,  
    and the covariance matrix of each pair of elements

    Args:
        data: a list of dicts with CIFs
    Return:
        write element histogram in elem_histo file
        write element-wise covariance matrix in elem_matrix file
    '''
    # define the indice by call element_indice function
    element_indice()

    # initialize two dicts
    elem_histo = {}
    elem_matrix = {}
    for elem in periodic_table:
        elem_histo[elem] = 0
        elem_matrix[elem] = {}
        for elem2 in periodic_table:
            elem_matrix[elem][elem2] = 0

    for d in data:
        elements = Structure.from_str(d['cif'], fmt='cif').symbol_set
        for elem in elements:
            elem_histo[elem] += 1
        for elem_pair in combinations(elements, 2):
            elem_matrix[elem_pair[0]][elem_pair[1]] += 1

    for elem in periodic_table:
        for elem2 in periodic_table:
            if elem_matrix[elem][elem2] == 0 and elem != elem2:
                elem_matrix[elem][elem2] = elem_matrix[elem2][elem]

    elem_histo = {elem:count for elem,count in elem_histo.items() if count != 0}
    elem_histo = pd.DataFrame([elem_histo])
    elem_histo = elem_histo[(elem_histo != 0)]
    elem_histo.T.to_csv('elem_histo', sep=' ')

    #
    elem_matrix = pd.DataFrame.from_dict(elem_matrix)
    # remove columns and rows with all zero
    elem_matrix = elem_matrix.loc[:, (elem_matrix != 0).any(axis=0)] # column
    elem_matrix = elem_matrix[(elem_matrix != 0).any(axis=1)]   # row
    elem_matrix.to_csv('elem_matrix', sep=' ')

    return


def elements_selection(data, elem_list, mode='include'):
    '''
    Select a subset contains or not contain certain elements.

    Args:
        data: a list of dicts with CIFs
        elem_list: a list of the elements of interest or no-interest
        mode:
            include: select structures have elements in elem_list
            exclude: drop structures have elements in elem_list
            consist: select structures made up of elements in elem_list
    Return:
        A new dataset after selection
    '''
    
    new_data = []
    
    for d in data[:]:
        elements = Structure.from_str(d['cif'], fmt='cif').symbol_set
        if mode == 'include':
            if set(elem_list).isdisjoint(elements):
                continue
                #data.remove(d)
        elif mode == 'exclude':
            if not set(elem_list).isdisjoint(elements):
                continue
                #data.remove(d)
        elif mode == 'consist':
            if not set(elements).issubset(elem_list):
                continue
                #data.remove(d)
                #print("Removed", d['task_id'], set(elements))
        new_data.append(d)
    return new_data


def similarity_matrix(input_file='dist_matrix', normalize='inverse', order='pt_number'):
    '''
    Read the original file from the below reference, and convert into a dictionary

    Hautier, G., et al. (2011). 
    "Data mined ionic substitutions for the discovery of new compounds." 
    Inorganic Chemistry 50(2): 656-663.

    Args:
        input_file: the preprocessed supporting txt file (from the above paper)
                using the shell script below
        normalize: method for value normalization
            bound: all value divide by 20 (based on the maximum value)
            log: all value got log10
            inverse: 1 / values
        order: the order of element in the matrix
            pt_number: the order in periodic table, i.e. atomic number
                    typically for calculation purpose
            pettifor: typically for visualization purpose, 
     Return:
        a pandas dataframe of the similarity matrix
    ====================================================
    #!/bin/sh
    outfile=dist_matrix
    > $outfile
    for i in `seq 78` # up to Bi and remove nobel gas
    do
        e1=`sed -n "$i p" mendlev` # element symbols in Pettifor order
        p1=`sed -n "$i p" Pettifor` # element symbol + valency in Pettifor order
        for j in `seq 78`
        do
            e2=`sed -n "$j p" mendlev`
            p2=`sed -n "$j p" Pettifor`
            if [ $i -gt $j ]
            then
                r=`grep $p1 ic102031h_si_001.txt | grep $p2`
                if [ -z "$r" ]
                then
                    grep -w $e1 ic102031h_si_001.txt | grep -w $e2 | head -n 1 >> $outfile
                else
                    echo $r >> $outfile
                fi
            fi
        done
    done
    sed -i 's/:/ /g' $outfile # make the valency a seperate column
    =========================================================
    '''
    # define the indice by call element_indice function
    element_indice()

    # note that the index_col is after the use of selection
    d = pd.read_csv(input_file, sep=' ', index_col=[0,1], usecols=[0,2,4])
    # make it a two dimensional matrix
    d = d.unstack()
    # drop the multilevel when ustack
    d.columns = d.columns.droplevel()

    if order == 'pt_number':
        index = periodic_table_78
    elif order == 'pettifor':
        index = pettifor
    # reindex, i.e. change the order of column and rows to Pettifor order  
    d = d.fillna(0)
    d = d.reindex(columns=index, index=index, fill_value=0) 
    d = d + d.transpose()
    # the maximum similarity number is 18.6, set the same element to 20, a bit arbitary
    np.fill_diagonal(d.values, 20)
    # then fill the zeros
    d.loc[:,'Pm'] = d.loc[:,'Sm'] # column, same as:  d['Pm'] = d['Sm']
    d.loc['Pm',:] = d.loc['Sm',:] # row
    # other zero mean very dissimilar, so set to a very small value
    d.replace(0, 0.1, inplace=True)

    if normalize == 'bound':
        d = d / 20
    elif normalize == 'log':
        d = np.log10(d)
    elif normalize == 'inverse':
        d = 1 / d
        np.fill_diagonal(d.values, 0)
    else:
        print('normalization method not supported')

    return d


def bonding_type(structure):
    '''
    Calculate bonding in a given structure.

    Args:
        structure: a pymatgen structure
    Return:
        A list of atomic pairs which form bonding
    '''
    nn = CrystalNN()
    bond_elem_list = []
    bond_num_list = []
    for i in list(range(len(structure))):
        site1 = structure[i].species_string
        num1 = structure[i].specie.number
        for neigh in nn.get_nn_info(structure, i):
            bond_elem_list.append(' '.join(sorted([site1, neigh['site'].species_string])))
            bond_num_list.append(' '.join(list(map(str, sorted([num1, neigh['site'].specie.number])))))

    bond_elem_list = list(set(bond_elem_list))
    bond_num_list = list(set(bond_num_list))
   
    return bond_elem_list, bond_num_list
            

def bonding_matrix(data):
    '''
    Convert the atomic pair of bonding into a matrix representation
    Remove the elements that don't exsit, then flatten it.

    Args:
        data: a list of dicts with CIFs
    Return:
        A 2D vector, first dimension is number of samples.
    '''
    # define the indice by call element_indice function
    element_indice()
    
    all_bond_matrix = []
    num_elements = len(periodic_table_78)
    zeor_bond_matrix = pd.DataFrame(np.zeros((num_elements, num_elements)), 
                                    index=periodic_table_78, columns=periodic_table_78)
    for d in data:
        bond_matrix = zeor_bond_matrix
        bond_list = d['bond_elem_list']
        for bond in bond_list:
            elem1, elem2 = bond.split()
            bond_matrix.loc[elem1, elem2] = 1
            bond_matrix.loc[elem2, elem1] = 1
        all_bond_matrix.append(bond_matrix.values.flatten())
    # HERE NEED TO DELETE ZERO ROW/COLUMN AND THEN FLATTEN
    return np.stack(all_bond_matrix)



def composition_hist(structure, 
                     neighbours,
                     order='Z',
                     normed=True,
                     rdf_type = 'grid',
                     return_sparse = False,
                     ):
    """
    Generate a 2D GRID-like histogram for composition, with composition grouped by increasing neighbour distance.

    Parameters
    ----------

    structure : PyMatGen Structure
    neighbours : list
        List of neighbours generated using `extendRDF.get_pairwise_distances`
    order : str
        Order to place elements in final representation. One of:
            - 'Z' (atomic number order)
            - 'pettifor' (pettifor order)
            - 'modified_pettifor' (modified and updated pettifor)

    normed : bool, default True
        Whether to normalise the composition of each shell to 1
    rdf_type : str, default 'grid'
        Whether to return a 1D histogram ('simple') or 2D histogram ('grid')
        of composition
    return_sparse: bool, default False
        If True, return a scipy.sparse.coo_matrix instead of NumPy array to save memory.
    """

    assert rdf_type == 'simple' or rdf_type == 'grid', "rdf_type must be one of `simple` (1D) or `grid` (2D). "

    # Set up vectors for composition and shell_number
    # Only treat 103 elements currently (using 0 for unknown elements)
    max_elements = 103
    max_shell = max([len(i) for i in neighbours])

    # First compute a 2D GRID-like representation of composition
    comp_grid = np.zeros((max_shell, max_elements+1))

    for nsite, site in enumerate(neighbours):
        for nshell, shell in enumerate(site):
            elem_fracs = shell.species.get_el_amt_dict()
            for elem in shell.species.elements:
                # Get index based on Z
                elem_index = elem.Z
                # Substitute Z for non-standard elements, which have a nonsense Z otherwise
                if elem.Z > max_elements or elem.Z < 0:
                    if elem.symbol == 'D':
                        elem_index = 1
                    elif elem.symbol == 'M':
                        # Some ICSD entries contain 'M'...?
                        elem_index = 0
                    else:
                        # Unknown element
                        warnings.warn(f"Unknown element type {elem.symbol} - assigning to column 0", UserWarning)
                        elem_index = 0
                    
                # Overall value is occupancy of site * fraction of element on neighbouring site
                # If all sites are fully occupied, this will just add 1
                comp_grid[nshell, elem_index] += elem_fracs[elem.symbol] * structure[nsite].species.num_atoms

    if rdf_type == 'simple':
        comp_grid = comp_grid.sum(axis=0)

    if normed:
        if rdf_type == 'grid':
            comp_grid = comp_grid / np.tile(comp_grid.sum(axis=-1).reshape(-1,1), comp_grid.shape[-1])
        elif rdf_type == 'simple':
            comp_grid = comp_grid / comp_grid.sum()

    if return_sparse:
        comp_grid = coo_matrix(comp_grid)

    return comp_grid


""" Grouped Representations of Interatomic Distances (GRID)

Package to compute, analyse and train ML models using the GRID descriptor
for crystal structures.


"""

__version__ = "0.1.3"


__all__ = ['composition',
           'data_explore',
           'data_io',
           'data_prepare',
           'earth_mover_distance',
           'extendRDF',
           'otherRDFs',
           'train',
           'visualization',
           ]

from . import composition
from . import data_explore
from . import data_io
from . import data_prepare
from . import earth_mover_distance
from . import extendRDF
from . import otherRDFs
from . import train
from . import visualization


""" Train machine learning models using GRID and RDF descriptors.

Utility functions to simplify training of general ML models using
GRID, composition and different dissimilarity measures (including
Earth mover's distance).
"""
import os
import sys
import json
import argparse
import numpy as np
import pandas as pd
try:
    from pymatgen import Structure
except ImportError:
    from pymatgen.core.structure import Structure
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import Lasso, ElasticNet
from sklearn.svm import SVR, SVC
from sklearn.neighbors import KNeighborsRegressor
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.model_selection import learning_curve, GridSearchCV
import sklearn.model_selection
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn import metrics
from pyemd import emd

from .data_explore import rdf_trim, rdf_flatten, batch_lattice
from .composition import composition_one_hot, bonding_matrix
from .visualization import calc_obs_vs_pred, binarize_output, n_best_middle_worst
#from .extendRDF import shell_similarity
from .data_io import rdf_read, rdf_read_parallel, shell_similarity_read
from .misc import int_or_str


# Manually create 'train_test_split' so it can be changed to 2D version if needed.
train_test_split = sklearn.model_selection.train_test_split

def train_test_split_2D(X_data, y_data, test_size, random_state=1):
    """ Modified version of train_test_split designed to split a square nxn matrix correctly. """
    
    print('Using 2D split')
    # Turn things (back?) into a DataFrame so we retain the indices
    full_X = pd.DataFrame(X_data)
    full_y = pd.DataFrame(y_data)
    
    X_train, X_test, y_train, y_test = \
        sklearn.model_selection.train_test_split(full_X, full_y, test_size=test_size, random_state=random_state, shuffle=True)
    
    #print(X_test.index)
     
    # Check we aren't leaking anything
    assert len(X_train.index.intersection(X_test.index)) == 0
    assert len(y_train.index.intersection(y_test.index)) == 0
        
    assert len(X_train.index.intersection(X_train.columns)) == len(X_train)
    assert len(X_test.index.intersection(X_test.columns)) == len(X_test)
    
    # To use precomputed distances correctly, we need a rectangular
    # "X_test" set with distances to the items in "X_train"
    X_test_rect = full_X.loc[X_test.index, X_train.columns]
    assert X_test_rect.shape == (X_test.shape[0], X_train.shape[1])    
    
    X_test_rect = X_test_rect.reindex(X_train.index, axis=1).to_numpy()        
    X_train = X_train.reindex(X_train.index, axis=1).to_numpy()
    X_test = X_test.reindex(X_test.index, axis=1).to_numpy()
    
    # Check we still have zero diagonals for distance
    assert (X_train.diagonal() == 0).all()
    assert (X_test.diagonal() == 0).all()
    

    
    
    # If only one training column, we need to call to_numpy differently
    # to avoid including indices in the output
    if y_train.shape[1] > 1:
        y_train = y_train.to_numpy()
        y_test = y_test.to_numpy()
    else:
        y_train = y_train[0].to_numpy()
        y_test = y_test[0].to_numpy()

    return X_train, X_test, X_test_rect, y_train, y_test
    
def calc_obs_vs_pred_2D(funct, X_data, y_data, test_size, outdir='./'):
    '''
    The observation vs prediction plot for 2D X data (e.g. similarity matrix)
    '''
    X_train, X_test, X_test_rect, y_train, y_test = \
            train_test_split_2D(X_data, y_data, test_size=test_size, random_state=1)
    
    
    
    funct.fit(X_train, y_train)
    y_pred = funct.predict(X_test_rect)
    y_pred_train = funct.predict(X_train)
    
    np.savetxt(os.path.normpath(os.path.join(outdir, 'test.' + str(test_size))), 
                np.stack([y_test, y_pred]).transpose(), 
                delimiter=' ', fmt='%.3f')
    np.savetxt(os.path.normpath(os.path.join(outdir, 'train.' + str(test_size))), 
                np.stack([y_train, y_pred_train]).transpose(), 
                delimiter=' ', fmt='%.3f')

def krr_grid_search(alpha, gamma, X_data, y_data, test_size=0.2):
    kr = GridSearchCV( KernelRidge(),
                    scoring='neg_mean_absolute_error',
                    param_grid=[#{'kernel': ['rbf'], 'alpha': alpha, 'gamma': gamma},
                                {'kernel': ['linear'], 'alpha': alpha}]
                     )
    X_train, X_test, y_train, y_test = \
        train_test_split(X_data, y_data, test_size=test_size, random_state=1) 
    kr = kr.fit(X_train, y_train)
    return kr.best_score_ , kr.best_params_, kr.cv_results_


def svr_grid_search(gamma, C, X_data, y_data, test_size=0.2):
    svr = GridSearchCV( SVR(),
                    scoring='neg_mean_absolute_error',
                    param_grid=[{'kernel': ['rbf'], 'gamma': gamma, 'C': C},
                                {'kernel': ['linear'], 'C': C}]
                      )
    X_train, X_test, y_train, y_test = \
        train_test_split(X_data, y_data, test_size=test_size, random_state=1) 
    svr = svr.fit(X_train, y_train)
    return svr.best_score_ , svr.best_params_ #, svr.cv_results_


def calc_learning_curve(funct, X_data, y_data, test_size=0.1, procs=1, output_dir = '.'):
    #X_train, X_test, y_train, y_test = \
    #    train_test_split_2D(X_data, y_data, test_size=2, random_state=None)
        
    ##########
    #
    # If the data are passed straight to learning_curve, the MAE ends up
    # twice as big as if the data are permuted first (taking 2D correspondence into
    # account if a distance matrix is supplied). As such, we permute the data first
    #
    # This *probably* occurs because learning_curve performs the splitting for 
    # k-fold validation before 'shuffling' the data. In this case, I expect that
    # some folds end up with very similar data and some with very dissimilar data
    # (intuitive if MP IDs increase with increasing complexity over time).
    
    # Gives 20GPa:
    orders = np.random.permutation(np.arange(len(X_data)))
    # Gives 47 GPa:
    #orders = np.arange(len(X_data))
    # Give 46 GPa:
    #orders = np.hstack([np.arange(50, len(X_data)), np.arange(50)])
    
    if isinstance(funct, KNeighborsRegressor):
        # Assume we are using a distance matrix - keep row-column correspondence
        X_ = X_data[orders][:, orders]
    else:
        X_ = X_data[orders]
    y_ = y_data[orders]

    train_sizes, train_scores, test_scores = \
        learning_curve(estimator=funct, X=X_, y=y_, 
                        train_sizes=np.linspace(0.1, 1.0, 10),
                        #train_sizes = np.arange(0.1, 1+test_size - 0.01, test_size),
                        scoring='neg_mean_absolute_error',
                        shuffle=True,
                        verbose=0,
                        cv=5, n_jobs=procs) 

    print("Training scores:")
    print(train_scores)
    print("Testing scores")
    print(test_scores)
                        
    train_mean = np.mean(train_scores, axis=1) *-1
    train_std = np.std(train_scores, axis=1)
    test_mean = np.mean(test_scores, axis=1) * -1
    test_std = np.std(test_scores, axis=1)

    learning_curve_results = np.stack([ train_mean, train_std, test_mean, test_std ])
    learning_curve_results = learning_curve_results.transpose()

    np.savetxt(os.path.normpath(os.path.join(output_dir, 'learning_curve')), learning_curve_results, delimiter=' ', fmt='%.3f')

    print("{:11s} {:11s} {:11s} {:11s} {:11s}".format("Train size", "<Train>", "std(Train)", "<Test>", "std(Test)"))
    for i, result in enumerate(learning_curve_results):
        print('{0:11d} {1:10.3f} {2:10.3f} {3:10.3f} {4:10.3f}'.format(train_sizes[i], *result))
    

def emd_of_two_compositions(y_test, y_pred, pettifor_index=True):
    '''
    Calcalate the earth mover's distance of two compositions, the composition should be 
    a 78-element vector, as the similarity_matrix is 78x78 matrix in the order of  
    atom numbers

    Args:
        y_test: test data with dimension n*78, the second dimension is by default in the  
            pettifor order, see 'composition_one_hot' function in data_explore.py
        y_pred: prediction data
        pettifor_index: whether transform the element order from the peroidic table
            number to Pettifor number
    Return:

    '''
    dist = []
    elem_similarity_file = os.path.join(sys.path[0], 'similarity_matrix.csv')
    dist_matrix = pd.read_csv(elem_similarity_file, index_col='ionA')
    dist_matrix = 1 / (np.log10(1 / dist_matrix + 1))

    if pettifor_index:
        pettifor = ['Cs', 'Rb', 'K', 'Na', 'Li', 'Ba', 'Sr', 'Ca', 'Yb', 'Eu', 'Y',  'Sc', 'Lu', 'Tm', 'Er', 'Ho', 
            'Dy', 'Tb', 'Gd', 'Sm', 'Pm', 'Nd', 'Pr', 'Ce', 'La', 'Zr', 'Hf', 'Ti', 'Nb', 'Ta', 'V',  'Mo', 
            'W',  'Cr', 'Tc', 'Re', 'Mn', 'Fe', 'Os', 'Ru', 'Co', 'Ir', 'Rh', 'Ni', 'Pt', 'Pd', 'Au', 'Ag', 
            'Cu', 'Mg', 'Hg', 'Cd', 'Zn', 'Be', 'Tl', 'In', 'Al', 'Ga', 'Pb', 'Sn', 'Ge', 'Si', 'B',  'Bi', 
            'Sb', 'As', 'P',  'Te', 'Se', 'S', 'C', 'I', 'Br', 'Cl', 'N', 'O', 'F', 'H']
        dist_matrix = dist_matrix.reindex(columns=pettifor, index=pettifor) 
    dist_matrix = dist_matrix.values

    # the earth mover's distance package need order C and float64
    dist_matrix = dist_matrix.copy(order='C')
    for y_t, y_p in zip(y_test, y_pred):
        dist.append(emd(y_t.astype('float64'), y_p.astype('float64'), 
                    dist_matrix.astype('float64')))
    
    #print(len(y_pred), np.count_nonzero(np.array(dist)))    
    return np.stack(dist)


def bond_to_atom(data, nelem=78):
    '''
    Convert atomic pairs bonding into the exsitance of elements

    Args:
        Data: y_data of flatten matrix
        nelem: number of elements in the periodic table, determines the shape
            of the bonding matrix
    Return:
        One-hot vector of elements
    '''
    new_data = []
    for y in data:
        y = np.reshape(y, (nelem, nelem))
        # make all the non-zero values to 1
        # NOTETHAT the sum method is only valid for a symmetric bonding matrix
        new_data.append(np.sum(y, axis=0).astype(bool).astype(int))
    return np.stack(new_data)
    

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Train machine learning algorithm',
                                    formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--rdf_dir', type=str, default='./',
                        help='The dir of all rdf files for training \n' +
                            '   default is current work dir'
                        )
    parser.add_argument('--input_file', type=str, 
                        default='../MP_modulus_datasets/MP_modulus.json',
                        help='files contain target data \n' +
                            '   default is ../MP_modulus.json'
                        )
    parser.add_argument('--input_features', type=str, default='extended_rdf',
                        help='features used for machine learning: \n' +
                            '   extended_rdf  \n' +
                            '   shell_similarity: shell-wise similarity values \n' +
                            '   lattice_abc: a, b, c, alpha, beta, gamma \n' +
                            '   lattice_matrix: a 3x3 matrix of lattice \n' +
                            '   formula:  \n' + 
                            '   composition:  \n' +
                            '   distance_matrix: precomputed pair-wise distances (e.g. EMD) \n' +
                            ' '
                        )
    parser.add_argument('--output_dir', type=str, default = './',
                        help='directory to put output files'
                        )
    parser.add_argument('--funct', type=str, default='krr',
                        help='which function is used, currently support: \n' +
                            '   krr(default, Kernel Ridge Regression) \n' +
                            '   svm(Support Vector Machine) \n' +
                            '   rf(random forest) \n' +
                            '   lasso: linear model with L1 regulation \n' +
                            '   elastic_net: both L1 and L2 regulation are used \n' + 
                            '   knn_reg: KNeighborsRegressor'
                        )
    parser.add_argument('--target', type=str, default='bulk_modulus',
                        help='which properteis as target, currently support: \n' +
                            '   bulk_modulus \n' +
                            '   shear_modulus \n' +
                            '   density \n' +
                            '   composition: percentage of elements in each compound \n' +
                            '   type_of_elements: which types of elements in each compound \n' +
                            '   volume_per_atom \n' +
                            '   volume \n' +
                            '   space_group_number \n' +
                            '   average_coordination \n' +
                            '   number_of_atoms: in the unit cell \n' +
                            '   bonding_type:  \n' +
                            '   average_coordination:  \n' +
                            '   average_bond_length:  \n' +
                            '   bond_length_std:  \n' +
                            '   ave_bond_std:  \n' +
                            '   coord_num_std:  \n' +
                            '   num_sg_operation:  \n' +
                            '   number_of_species: number of atomic species of the compound \n' +
                            '   band_gap: band gap data contained in input_file (i.e. from materials project)'
                        )
    parser.add_argument('--log_target', type=bool, default=False, 
                        help='whether to take log of target before fitting (default False)'
                        )
    parser.add_argument('--task', type=str, default='test_size_depend',
                        help='one of the following: \n'+
                            '   test_size_depend: default, change the test size from 0.1 to 0.9 \n' +
                            '   obs_vs_pred: plot the observation vs prediction results for \n' +
                            '       training and test set respectively, observation go first \n' +
                            '   confusion_maxtrix: histogram for multt-label output \n' +
                            '   grid_search: cross validation grid search for meta parameters \n' +
                            '   randon_guess: difference between ground truth and random guess/constant \n' +
                            '   emd_visual:  \n' +
                            '   learning_curve: the calculated learning curve will be stored \n' +
                            '       in output_dir/learning_curve'
                        )
    parser.add_argument('--metrics', type=str, default='default',
                        help='which metrics is used, currently support: \n' +
                            '   default: mae for continuous data \n' +
                            '   mae: mean absolute error \n' +
                            '   mape: mean absolute pencentage error \n' +
                            '   emd: earth mover distance, i.e. wasserstein metrics'
                        )
    parser.add_argument('--trim', type=str, default='minimum',
                        help='the number of shells for RDF used as input feature\n' +
                            '   minimum: default, use the minimum number of shells \n' +
                            '   none: no trim, suitable for origin RDF  \n' +
                            '   or an integer number'
                        )
    parser.add_argument('--dist_matrix', type=str, default=None,
                        help = 'file containing pre-computed pairwise distances to use \n' +
                               'instead of Sklearn distance metric (e.g. EMD)'
                        )              
    parser.add_argument('--procs', type=int, default=1,
                        help='The number of processes to parallelise over, where implemented (default 1)'
                        )

    args = parser.parse_args()
    rdf_dir = args.rdf_dir
    input_file = args.input_file
    target = args.target
    metrics_method = args.metrics
    funct_name = args.funct
    task = args.task
    trim = int_or_str(args.trim)
    input_features = args.input_features.split()
    dist_matrix = args.dist_matrix
    output_dir = args.output_dir
    procs = args.procs

    # read the dataset from input file
    with open (input_file,'r') as f:
        data = json.load(f)

    # the following input features are prepared
    # if the task is to give randon guess (for comparsion with prediction)
    # then no input features is needed
    # otherwise the input features are calculated and comibined
    if task != 'random_guess':
        all_features = [
            'extended_rdf', #'shell_similarity', 
            'lattice_abc', 'lattice_matrix',
            'formula', 'composition',
            'distance_matrix',
        ]
        if not set(input_features).issubset(all_features):
            raise ValueError('Specified feature(s) are not available')

        X_data = None
        if 'extended_rdf' in input_features:
            if 'tar' in rdf_dir:
                all_rdf = rdf_read_tar(data, rdf_dir)
            else:
                if procs > 1:
                    all_rdf = rdf_read_parallel(data, rdf_dir, procs=procs)
                else:
                    all_rdf = rdf_read(data, rdf_dir)
        
            # make all the rdf same length for machine learning input
            all_rdf = rdf_trim(all_rdf, trim=trim)
            X_data = rdf_flatten(all_rdf)
        # if 'shell_similarity' in input_features:
        #     all_shell_simi = shell_similarity_read(data, rdf_dir)
        #     if X_data is None:
        #         X_data = all_shell_simi
        #     else: 
        #         X_data = np.hstack((X_data, all_shell_simi))
        if 'lattice_abc' in input_features:
            all_lattice = batch_lattice(data, method='abc')
            if X_data is None:
                X_data = all_lattice
            else:
                X_data = np.hstack((X_data, all_lattice))
        if 'lattice_matrix' in input_features:
            all_lattice = batch_lattice(data, method='matrix')
            if X_data is None:
                X_data = all_lattice
            else:
                X_data = np.hstack((X_data, all_lattice))
        if 'distance_matrix' in input_features:
            assert(dist_matrix is not None)
            # Change to 2D version of train_test_split (also in sub-packages)
            train_test_split_old = train_test_split
            train_test_split = train_test_split_2D
            
            calc_obs_vs_pred = calc_obs_vs_pred_2D
            print('Reading distance matrix ... ', end=''),
            X_data = pd.read_csv(dist_matrix, index_col=0)
            
            # Check that distance matrix matches data file for MP-ID
            assert len(X_data.index.intersection( [d['task_id'] for d in data] )) == len(X_data)
            print('Done')
            # Finally, throw away dataFrame indices
            X_data = X_data.to_numpy()
        
    # the following line save X_data for check, but this is rarely used
    #np.savetxt(os.path.join(output_dir, 'X_data'), X_data, delimiter=' ',fmt='%.3f')
    print("Shape of input feature array: ", X_data.shape)
    # target_type can be continuous categorical ordinal
    # or multi-cont, multi-cate, multi-ord
    target_type = None

    # calculate the target values
    if target == 'bulk_modulus':
        target_type = 'continuous'
        y_data = np.array([ x['elasticity.K_VRH'] for x in data ])
        if args.log_target:
            y_data = np.log10(y_data)

    elif target == 'shear_modulus':
        target_type = 'continuous'
        y_data = np.array([ x['elasticity.G_VRH'] for x in data ])
        if args.log_target:
            y_data = np.log10(y_data)
    elif target == 'density':
        target_type = 'continuous'
        y_data = np.array([ Structure.from_str(x['cif'], fmt='cif').density 
                            for x in data ])
    elif target == 'volume_per_atom':
        target_type = 'continuous'
        y_data = np.array([ Structure.from_str(x['cif'], fmt='cif').volume 
                            / len(Structure.from_str(x['cif'], fmt='cif')) 
                            for x in data ])
    elif target == 'volume':
        target_type = 'continuous'
        y_data = np.array([ Structure.from_str(x['cif'], fmt='cif').volume 
                            for x in data ])
    elif target == 'space_group_number':
        target_type = 'continuous'
        y_data = np.array([ Structure.from_str(x['cif'], fmt='cif').get_space_group_info()[1]
                            for x in data ])
    elif target == 'number_of_species':
        #target_type = 'ordinal'
        #y_data = np.array([ len(Structure.from_str(x['cif'], fmt='cif').symbol_set) 
        #                    for x in data ])
        target_type = 'categorical'
        y_data = [ str(len(Structure.from_str(x['cif'], fmt='cif').symbol_set)) 
                            for x in data ]
    elif target == 'number_of_atoms':
        target_type = 'ordinal'
        y_data = np.array([ len(Structure.from_str(x['cif'], fmt='cif')) 
                            for x in data ])
    elif target == 'composition':
        target_type = 'multi-cont'
        y_data, elem_symbols = composition_one_hot(data=data, method='percentage', index='pettifor')
    elif target == 'type_of_elements':
        target_type = 'multi-cate'
        y_data, elem_symbols = composition_one_hot(data=data, method='only_type', index='pettifor')
    elif target == 'bonding_type':
        target_type = 'multi-cate'
        y_data = bonding_matrix(data=data)
    elif target == 'average_coordination':
        target_type = 'continuous'
        y_data = np.array([ x['average_coordination'] for x in data ])
    elif target == 'average_bond_length':
        target_type = 'continuous'
        y_data = np.array([ x['average_bond_length'] for x in data ])
    elif target == 'bond_length_std':
        target_type = 'continuous'
        y_data = np.array([ x['bond_length_std'] for x in data ])
        if args.log_target:
            y_data = np.log10(y_data + 1)
    elif target == 'ave_bond_std':
        target_type = 'continuous'
        y_data = np.array([ x['ave_bond_std'] for x in data ])
    elif target == 'coord_num_std':
        target_type = 'continuous'
        y_data = np.array([ x['coord_num_std'] for x in data ])
    elif target == 'num_sg_operation':
        target_type = 'continuous'
        y_data = np.array([ x['num_sg_operation'] for x in data ])
    elif target == 'band_gap':
        assert 'band_gap' in data[0].keys()
        target_type = 'continuous'
        y_data = np.array([ x['band_gap'] for x in data ])
    else:
        print('This target is not support, please check help')
        exit()

    print('Setting up model ... ', end='')
    # select the machine learning algorithm
    if funct_name == 'krr':
        funct = KernelRidge(alpha=1.0)
    elif funct_name == 'svm':
        if target_type in ('categorical', 'multi-cate'):
            funct = SVC(kernel='linear')
        else:
            funct = SVR(kernel='linear')
    elif funct_name == 'lasso':
        funct = Lasso()
    elif funct_name == 'elastic_net':
        funct = ElasticNet(alpha=1.0, l1_ratio=0.7)
    elif funct_name == 'rf':
        if target_type in ('categorical', 'multi-cate'):
            funct = RandomForestClassifier(max_depth=2, random_state=0)
        else:
            funct = RandomForestRegressor(max_depth=2, random_state=0)
    elif funct_name == 'knn_reg':
        if dist_matrix is None:
            print('knn requires a pre-computed distance matrix')
            exit()
        funct = KNeighborsRegressor(n_neighbors=1, metric='precomputed', weights='distance')
    else:
        print('this algorithm is not supported, please check help')
        exit()

    print('Done. Using model ', funct.__class__)

    # use a 80/20 split except otherwise stated, e.g. in varying test_size
    test_size = 0.2
    if task == 'test_size_depend':
        for test_size in np.linspace(0.9, 0.1, 9):
            X_train, X_test, y_train, y_test = \
                train_test_split(X_data, y_data, test_size=test_size, random_state=1) 

            funct.fit(X_train, y_train)
            y_pred = funct.predict(X_test)

            if target_type == 'continuous':
                if metrics_method == 'default' or metrics_method == 'mae':
                    pred_acc = metrics.mean_absolute_error(y_test, y_pred)
                elif metrics_method == 'mape':
                    pred_acc = metrics.mean_absolute_percentage_error(y_test, y_pred)
                else:
                    print('This metrics is not support for continuous data')
            elif target_type == 'multi-cont':
                if metrics_method == 'default':
                    pred_acc = [ metrics.mean_absolute_error(y_test[:,i], y_pred[:,i])
                                for i in range(len(y_test[0])) ]
                elif metrics_method == 'emd':
                    pass
                else:
                    print('This metrics is not support for multi-continuous data')
            elif target_type == 'categorical'and target == 'number_of_species':
                # now only implemented for number of species
                #y_pred[np.where(y_pred > 5)] = 5
                #y_pred = list(map(str, list(np.int64(y_pred + 0.5))))
                #print(type(y_test), type(y_test[0]), y_test[0], type(y_pred), type(y_pred[0]), y_pred[0])
                #pred_acc = metrics.accuracy_score(y_test, y_pred)
                y_test = list(map(float, y_test))
                pred_acc = metrics.mean_absolute_error(y_test, y_pred)
            elif target_type == 'multi-cate':
                if target == 'bonding_type':
                    y_test = bond_to_atom(y_test)
                    np.where(y_pred > 0.2, y_pred, 0)
                    y_pred = bond_to_atom(y_pred)

                if metrics_method == 'default':
                    pred_acc = [round(metrics.coverage_error(y_test, y_pred), 3),
                            round(metrics.label_ranking_average_precision_score(y_test, y_pred), 3),
                            round(metrics.label_ranking_loss(y_test, y_pred), 3) ]
                elif metrics_method == 'emd':
                    #for nelem in range(2,8):
                    y_bin = binarize_output(y_test, y_pred, threshold=None, 
                                            nelem=None, save_to_file=False)
                    pred_acc = emd_of_two_compositions(y_test, y_bin).mean()
                    #print(test_size, nelem, pred_acc)
                else:
                    print('This metrics is not support for multi-label data')
            elif target_type == 'ordinal':
                y_pred = np.int64(y_pred + 0.5)
                if metrics_method == 'default' or metrics == 'mae':
                    pred_acc = metrics.mean_absolute_error(y_test, y_pred)
                elif metrics_method == 'mape':
                    pred_acc = metrics.mean_absolute_percentage_error(y_test, y_pred)
                else:
                    print('This metrics is not support for ordinal data')
            else:
                print('target type not supported')

            print('Training size {} ; Training samples {} ; Metrics {}'.format(
                    round(1-test_size, 3), int((1-test_size)*len(y_data)), pred_acc))
    
    elif task == 'obs_vs_pred':
        calc_obs_vs_pred(funct=funct, X_data=X_data, y_data=y_data, test_size=test_size,
                        outdir= output_dir)
    
    elif task == 'confusion_matrix':
        if target_type == 'multi-cate':
            X_train, X_test, y_train, y_test = \
                    train_test_split(X_data, y_data, test_size=test_size, random_state=1)
            funct.fit(X_train, y_train)
            y_pred = funct.predict(X_test)
            # if the prediction is decmical, use this to round up
            y_bin = binarize_output(y_test, y_pred, threshold=0.4, save_to_file=False)

            # Calcuate label-wise (typically element-wise) confustion matrix for each label. 
            # The output is reshaped into a 'number of types of lables' by 4 
            # in the order of true-positive false-positive false-negative true-negative
            cm = metrics.multilabel_confusion_matrix(y_test, y_bin)
            np.savetxt(os.path.normpath(os.path.join(output_dir, 'confusion_matrix_' + str(round(test_size,3)))), 
                        cm.reshape(len(y_pred[0]), 4), delimiter=' ', fmt='%.3f')
        elif target_type == 'categorical' and target == 'type_of_elements':
            # now only implemented for type of elements
            classes = list(map(str, list(range(1,6))))
            X_train, X_test, y_train, y_test = \
                    train_test_split(X_data, y_data, test_size=test_size, random_state=1)
            funct.fit(X_train, y_train)
            y_pred = funct.predict(X_test)
            y_pred[np.where(y_pred > 5)] = 5
            y_pred = list(map(str, list(np.int64(y_pred + 0.5))))
            cm = metrics.confusion_matrix(y_test, y_pred, labels=classes)
            np.savetxt(os.path.normpath(os.path.join(output_dir, 'confusion_matrix_multiclass'), cm, delimiter=' ', fmt='%.3f'))
        else:
            print('This target does not support confusion matrix')
    
    elif task == 'grid_search':
        if funct_name == 'krr':
            best_score, best_params, cv_results = \
                    krr_grid_search(alpha=np.logspace(-2, 2, 5), gamma=np.logspace(-2, 2, 5),
                                    X_data=X_data, y_data=y_data, test_size=test_size)
            print('best score {} ; best parameter {}'.format(best_score, best_params))
        elif funct_name == 'svm' and target_type not in ('categorical', 'multi-cate'):
            best_score, best_params, cv_results = \
                    svr_grid_search(gamma=np.logspace(-8, 1, 10), C=[1, 10, 100, 1000],
                                    X_data=X_data, y_data=y_data, test_size=test_size)
            print('best score {} ; best parameter {}'.format(best_score, best_params))
    
    elif task == 'learning_curve':
        calc_learning_curve(funct=funct, X_data=X_data, y_data=y_data, test_size=test_size, procs=procs)

    elif task == 'random_guess':
        if target == 'number_of_species':
            nclass = 5
            max_class = 3

            y_pred_const = np.full(len(y_data), max_class)
            pred_acc_const = metrics.mean_absolute_error(y_data, y_pred_const)
            y_pred_rand = np.int64(np.random.rand(len(y_data)) * nclass + 1)
            pred_acc_rand = metrics.mean_absolute_error(y_data, y_pred_rand)

            print('constant {} ; random {}'.format(pred_acc_const, pred_acc_rand))
        
        elif target == 'type_of_elements':
            nelem = 78
            y_pred_rand = []
            for i in range(len(y_data)):
                y_rand = np.zeros(nelem)
                for j in np.random.randint(0, high=nelem, size=3):
                    y_rand[j] = 1
                y_pred_rand.append(y_rand)

            print(emd_of_two_compositions(y_data, y_pred_rand, pettifor_index=True).mean())

    elif task == 'emd_visual':
        X_train, X_test, y_train, y_test = \
                train_test_split(X_data, y_data, test_size=test_size, random_state=1)
        funct.fit(X_train, y_train)
        y_pred = funct.predict(X_test)
        
        for threshold in np.linspace(0.2, 0.6, 5):
            y_bin = binarize_output(y_test, y_pred, threshold=threshold, save_to_file=False)
            pred_acc = emd_of_two_compositions(y_test, y_bin)
            np.savetxt(os.path.normpath(os.path.join(output_dir, 'dist_histo_' + str(threshold))), pred_acc, delimiter=' ', fmt='%.3f')
            print(pred_acc.mean())
            
            large_samples, middle_samples, small_samples = \
                    n_best_middle_worst(y_test, y_bin, metrics_values=pred_acc, n_visual=100)
            np.savetxt(os.path.normpath(os.path.join(output_dir, 'large_sample_' + str(threshold))), large_samples,
                        delimiter=' ', fmt='%.3f')
            np.savetxt(os.path.normpath(os.path.join(output_dir, 'middle_sample_' + str(threshold))), middle_samples, 
                        delimiter=' ', fmt='%.3f')            
            np.savetxt(os.path.normpath(os.path.join(output_dir, 'small_sample_' + str(threshold))), small_samples, 
                        delimiter=' ', fmt='%.3f')

    else:
        print('This task is not supported')





"""
Miscellaneous utility functions to handle structures and GRID descriptions.

"""

import numpy as np
import pandas as pd
import json
import time
import gzip
import argparse
import os
from tqdm import tqdm
from pyemd import emd, emd_with_flow
try:
    # pymatgen < v2022
    from pymatgen import Structure
except ImportError:
    # pymatgen >= v2022
    from pymatgen.core.structure import Structure
from pymatgen.analysis.local_env import CrystalNN
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer 
try:
    from matminer.featurizers.structure import StructuralComplexity
except:
    print('matminer is not installed, cannot calculate structural complexity')

from .extendRDF import rdf_histo, rdf_kde, get_rdf_and_atoms#, shell_similarity
from .composition import elements_count, bonding_type
from .otherRDFs import origin_rdf_histo


# def batch_shell_similarity(all_rdf, method='append'):
#     '''
#     Calculate shell similarity for multiple RDFs
    
#     Args:
#         all_rdf: multiple RDFs with the same length
#         method: how shell similarity is used in the X input
#             append: append the similarity values after rdfs
#             only: only use similarity values as the X input
#     Return:
#         RDF with shell similarity or similarity only
#     '''
#     rdf_shell_simi = []
#     for rdf in all_rdf:
#         if method == 'append':
#             rdf_shell_simi.append(np.append(rdf, shell_similarity(rdf)))
#         elif method == 'only':
#             rdf_shell_simi.append(shell_similarity(rdf))
#     return np.stack(rdf_shell_simi)


def rdf_trim(all_rdf, trim='minimum'):
    '''
    Trim the rdfs to the same length for kernel methods training

    Args:
        all_rdf: a list of np array (rdfs) with different length
        trim: must be one of 'no', 'minimum' or an integer
            no: no trim when the RDF already have the same length, 
            minimum: all rdf trim to the value of the smallest rdf
            integer value: if a value is given, all rdf longer than 
                this value will be trimmed, short than this value will  
                add 0.000 to the end
    Return:
        a two dimensional matrix, first dimension is the number of
        samples, and the second dimension is the flatten 1D rdf    
    '''
    rdf_lens = []
    for rdf in all_rdf:
        rdf_lens.append(len(rdf))

    if trim == 'minimum':
        min_len = np.array(rdf_lens).min()
        all_rdf = [ x[:min_len] for x in all_rdf]
    elif isinstance(trim, int):
        for i, rdf_len in enumerate(rdf_lens):
            if rdf_len < trim:
                b = trim - rdf_len
                if b != 0:
                    if len(all_rdf[0].shape) == 2:
                        nbins = len(all_rdf[i][0])
                        all_rdf[i] = np.append( all_rdf[i], [[0.0] * nbins] * b, axis=0 )
                    elif len(all_rdf[0].shape) == 1:
                        all_rdf[i] = np.append( all_rdf[i], [0.0] * b )
            else:
                all_rdf[i] = all_rdf[i][:trim]

#            print(len(all_rdf[i]))
    elif trim == 'none':
        pass 
    else:
        print('wrong value provided for trim') 

    return np.stack(all_rdf)


def rdf_flatten(all_rdf):
    '''
    If the rdf is not 1d, make it 1d for machine learning input
    This is done after trim so np.stack can be used

    Args:
        all_rdf: a 2D or 3D np array (rdfs) with the same length, 
            with the first dimension be number of samples
    Return:
        a 2D np array of rdfs
    '''
    if len(all_rdf[0].shape) == 2:
        all_rdf = [ x.flatten() for x in all_rdf]
    return np.stack(all_rdf)


def batch_lattice(data, method='matrix'):
    '''
    Calculate lattice parameters of each structure

    Args:
        data: input data from Materials Project
        method:
            abc: binned rdf, equal to the 'uniform' method in kde,  
                note that the 
            matrix: use kernel density estimation to give a smooth curve
    Return:
        lattice parameters
    '''
    all_lattice = []
    for d in data:
        struct = Structure.from_str(d['cif'], fmt='cif')
        if method == 'matrix':
            all_lattice.append(struct.lattice.matrix.reshape(9))
        elif method == 'abc':
            all_lattice.append(np.array(struct.lattice.abc + struct.lattice.angles))
    return np.stack(all_lattice)


def bond_length_statis(structure): 
    '''
    Calculated some bond length statistics values in the structure
    the nearest atoms pairs are determined using pymatgen CrystalNN module

    Currently the range of bond length has some problem for structure with
    no bonding, i.e. empty np array

    Args:
        structrue: pymatgen structure
    Return:
        the mean value and standard deviation and all the cation-anion
        bond length in the crystal structure
    '''
    nn = CrystalNN()
    bond_len = []
    for j in range(len(structure)): 
        for i in nn.get_nn_info(structure, j):       
            bond_len.append(structure[j].distance(structure[i['site_index']]))
    bond_len = np.array(bond_len)
    return bond_len.mean(), bond_len.std()#, bond_len.ptp()


def average_coordination(structure):
    '''
    Calculation of average coordination number over every site in a structure
    using Vorini method

    Args:
        structure: pymatgen structure
    Return:
        Average coordination number
    '''
    nn = CrystalNN()
    ave_coord_num = []
    for atom_site in range(len(structure)):
        ave_coord_num.append(len(nn.get_nn_info(structure, atom_site)))
    return np.array(ave_coord_num).mean()


def bond_stat_per_site(structure):
    '''
    Get the bond length standard deviation of each site in the unit cell, the average
    Get the coordination number of each site, then standard deviation

    Args:
        structure: pymatgen structure
    Return:
        Average of bond length std of each site
        Standard deviation of coordination number
    '''
    nn = CrystalNN()
    coord_num = []
    bond_len_std = []
    for atom_site in range(len(structure)):
        struct_nn = nn.get_nn_info(structure, atom_site)
        coord_num.append(len(struct_nn))
        bond_len = []
        for i in struct_nn:       
            bond_len.append(structure[atom_site].distance(structure[i['site_index']]))
        bond_len_std.append(np.array(bond_len).std())
    return round(np.array(bond_len_std).mean(), 3), \
            round(np.array(coord_num).std(), 3)


def num_of_shells(data, dir):
    '''
    Calculate the number of nearest neighbor shells in RDF for each compound,  
    some other properties are also given for data analysis purpose

    Args:
        data: the bulk modulus data and structure from Materials Project
        dir: the dir contains the calculated RDFs
    Return:
        a list of properties, will change according to requrest
    '''
    #sc = StructuralComplexity()
    results = []
    for d in data:
        struct = Structure.from_str(d['cif'], fmt='cif')
        #complex_atom, complex_cell = sc.featurize(struct)
        '''num_shell = sum(1 for line in open(dir + '/' + d['task_id']))

        with open(dir + '/' + d['task_id']) as f:
            rdf = np.loadtxt(f, delimiter=' ')
            num_rdf_bar = np.count_nonzero(rdf[:30])
        '''
        bond_mean, bond_std = bond_length_statis(struct)

        results.append(np.array([
            #d['elasticity.K_Voigt'],           # bulk modulus
            #num_shell,                          # number of RDF shells
            #num_rdf_bar,                        # non-zeors in the first 30 RDF shell
            #struct.volume/len(struct),          # volume per atom
            struct.volume,                      # volume
            struct.get_space_group_info()[1],   # space group number
            bond_mean,
            bond_std,
            #,     # average coordination nunmber
            #struct.density,                     # density in g/cm^3
            len(struct),                        # number of atoms
            len(struct.symbol_set),             # number of types of elements
            #sc.featurize(struct)[0],    # structural complexcity per atom
        ]))

    return np.stack(results)


def rdf_value_stat(data, dir):
    '''
    Remove all the 0 values in all RDF and put all other values in one list.  
    To get a statistics intuition of the RDF values distribution histogram.

    Args:
        dir: the dir contains the calculated RDFs
    Return:
        a list of all non-zero RDF values
    '''
    for d in data:
        struct = Structure.from_str(d['cif'], fmt='cif')
        len(struct)
        outfile = d['task_id']
    return


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Data explore',
                                    formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--input_file', type=str, default='../MP_modulus.json',
                        help='the bulk modulus and structure from Materials Project')
    parser.add_argument('--output_file', type=str, default='rdf_similarity',
                        help='currently for rdf similarity')
    parser.add_argument('--rdf_dir', type=str, default='./',
                        help='dir has all the rdf files')
    parser.add_argument('--task', type=str, default='num_shell',
                        help='which property to be calculated: \n' +
                            '   num_shell: number of RDF shells and other propreties \n' +
                            '   extend_rdf_bin: binned extend RDF of all the CIFs \n' +
                            '   extend_rdf_kde: kernel density estimation RDF \n' + 
                            '   origin_rdf: calcualte vanilla RDF of all the CIFs \n' + 
                            '   composition: element-wise statistics of all compositions \n' +
                            '   bonding_type: \n' +
                            '   shell_similarity: \n' +
                            '   '
                      )


    args = parser.parse_args()
    input_file = args.input_file
    output_file = args.output_file
    rdf_dir = args.rdf_dir
    task = args.task

    max_dist = args.max_dist

    with open(input_file,'r') as f:
        data = json.load(f)

    if task == 'num_shell':
        results = num_of_shells(data=data, dir=rdf_dir)
        outfile = '../num_shell'
        np.savetxt(outfile, results, delimiter=' ', fmt='%.3f')

    elif task == 'composition':
        elements_count(data)


    elif task == 'bonding_type':
        for d in data:
            struct = Structure.from_str(d['cif'], fmt='cif')
            d['bond_elem_list'], d['bond_num_list'] = bonding_type(struct)
        with open(input_file.replace('.json','_with_bond.json'), 'w') as f:
            json.dump(data, f, indent=1)

    elif task == 'bond_stat_per_site':
        for d in data:
            struct = Structure.from_str(d['cif'], fmt='cif')
            d['ave_bond_std'], d['coord_num_std'] = bond_stat_per_site(struct)
            d['num_sg_operation'] = len(SpacegroupAnalyzer(struct).get_space_group_operations())
        with open(input_file.replace('v7','v8'), 'w') as f:
            json.dump(data, f, indent=1)
            
    elif task == 'shell_similarity':
        for i, d in enumerate(data):
            if ( i % 100 ) == 0:
                print(i) 
            with open(rdf_dir + '/' + d['task_id']) as f:
                rdf = np.loadtxt(f, delimiter=' ')
                np.savetxt('../shell_similarity/' + d['task_id'], 
                            shell_similarity(rdf[:30]), 
                            delimiter=' ', fmt='%.3f')


    else:
        print('This task is not supported')




"""
Module to compute earth mover's distance (EMD) for different inputs.

Allows EMD comparison between the following features:
- Composition (as normalised vector)
- GRID histograms (mean of pairwise comparisons)
- RDF

The module also includes helper functions to compute dissimilarity matrices
between multiple input pairs, for later ML.
"""

import sys
import os
import json
import math
import numpy as np
import pandas as pd
import argparse
from tqdm import tqdm
try:
    from pymatgen import Structure
except ImportError:
    from pymatgen.core.structure import Structure   
try:
    from pymatgen import Lattice
except ImportError:
    from pymatgen.core.lattice import Lattice   
from pymatgen.analysis.structure_matcher import StructureMatcher
from scipy import spatial
from scipy.sparse import coo_matrix

import warnings

# Used for faster version of EMD
import numba as nb

# the wasserstein distance in scipy treats frequencies of each bin as a value, 
# and then builds the distributions from those values and computes the distance. 
# You can either simply pass the values that you create histograms from 
# or pass mid points of bins as values and frequencies as weights
# here the later method is used
from scipy.stats import wasserstein_distance

# generally the wasserstein distance in scipy is much faster than EMD package
# these two have been checked, they give same results
# so this emd is used only when
#   1. the flow between two distribution is needed
#   2. a distance matrix need to be defined, like composition similarity
from pyemd import emd, emd_with_flow



from .data_io import rdf_read, rdf_read_parallel
from .data_explore import rdf_trim, rdf_flatten
import gridrdf.composition
from .composition import composition_one_hot, element_indice
from .misc import int_or_str

def emd_formula_example():
    '''
    Test the Earth Mover's Distance (EMD) using similarity matrix 
    against the EMD in the literature
    https://github.com/lrcfmd/ElMD/
    '''
    
    try:
        from ElMD import ElMD
    except:
        print('Element-EMD module not installed')
        
    elem_emd = ElMD()
        
    comp1 = elem_emd._gen_vector('Li0.7Al0.3Ti1.7P3O12')
    comp2 = elem_emd._gen_vector('La0.57Li0.29TiO3')
    pettifor_emd = elem_emd._EMD(comp1, comp2)

    comp1_reindex = pd.DataFrame(comp1, index=gridrdf.composition.modified_pettifor).reindex(index=gridrdf.composition.pettifor)
    comp2_reindex = pd.DataFrame(comp2, index=gridrdf.composition.modified_pettifor).reindex(index=gridrdf.composition.pettifor)

    dist_matrix = pd.read_csv('similarity_matrix.csv', index_col='ionA').values
    dist_matrix = dist_matrix.copy(order='C')

    em = emd_with_flow(comp1_reindex.values[:,0], comp2_reindex.values[:,0], dist_matrix)
    simi_matrix_emd = em[0]
    emd_flow = pd.DataFrame(em[1], columns=gridrdf.composition.pettifor, index=gridrdf.composition.pettifor)
    emd_flow.replace(0, np.nan).to_csv('emd_flow.csv')


def find_same_structure(data):
    '''
    Find all pairs of structures that give exactly the same extended RDF
    in a dataset
    
    Args:
        data: json data containing structures information and MP-ID(typically CIFs)
    Return:
        A nested list of MP-IDs with same structures, grouped by space group number
    '''
    # group the structures (without composition information) using space group
    # first make a dict with keys are space group numbers
    sg_grouped_structs = {}
    for sg_num in range(230, 0, -1):
        sg_grouped_structs[sg_num] = {}
    
    for d in data:
        struct = Structure.from_str(d['cif'], fmt='cif')
        # remove the composition information by setting all elements to X
        # as currently extended RDF does not have composition information
        for elem in struct.symbol_set:
            struct[elem] = 'H'

        sg_num = struct.get_space_group_info()[1]
        sg_grouped_structs[sg_num][d['task_id']] = struct 
    
    sc = StructureMatcher(ltol=0.05, stol=0.05, scale=False)
    match_list = {}
    for sg_num, structs in sg_grouped_structs.items():
        print(sg_num, len(structs), flush=True)
        match_list[sg_num] = []

        while len(structs) != 0:
            struct_1 = list(structs.values())[0]
            same_struct_list = {}
            vol_per_atom_1 = struct_1.volume / struct_1.num_sites
            for task_id_2, struct_2 in structs.items():
                vol_per_atom_2 = struct_2.volume / struct_2.num_sites
                if abs(vol_per_atom_1 - vol_per_atom_2) < 1:
                    if sc.fit(struct_1, struct_2):
                        struct_2_latt = struct_2.get_primitive_structure().lattice.a
                        same_struct_list[task_id_2] = [vol_per_atom_2, struct_2_latt]
            if len(same_struct_list) > 1:
                match_list[sg_num].append(same_struct_list)
            for mp_id in same_struct_list:
                structs.pop(mp_id)

        '''
        if len(structs) > 0:
            match_list[sg_num] = []
            for i, (task_id_1, struct_1) in enumerate(tqdm(structs.items(), mininterval=60)):
                vol_per_atom_1 = struct_1.volume / struct_1.num_sites
                for j, (task_id_2, struct_2) in enumerate(structs.items()):
                    vol_per_atom_2 = struct_2.volume / struct_2.num_sites
                    if i < j and abs(vol_per_atom_1 - vol_per_atom_2) < 1:
                        if sc.fit(struct_1, struct_2):
                            match_list[sg_num].append([task_id_1, task_id_2])
                            print(task_id_1, task_id_2)
                            break
        '''
    return match_list


def find_same_rdf(all_rdf, data):
    '''
    Find all exactly the same extended RDF in a dataset
    
    Args:
        all_rdf: all the trimmed and flatten rdf
        data: json data containing structures information and MP-ID
    Return:
        A list of pairs of MP-IDs
    '''
    # make all_rdf a dict with keys of mp-ids
    rdfs = {}
    for i, d in enumerate(data):
        rdfs[d['task_id']] = all_rdf[i]
    
    match_list = []
    while len(rdfs) != 0:
        rdf_1 = list(rdfs.values())[0]
        same_rdf_list = []
        for task_id_2, rdf_2 in rdfs.items():
            if np.array_equal(rdf_1, rdf_2):
                same_rdf_list.append(task_id_2)
        if len(same_rdf_list) > 1:
            match_list.append(same_rdf_list)
        for mp_id in same_rdf_list:
            rdfs.pop(mp_id)
    return match_list


def analysis_emd_100():
    '''
    '''
    df = pd.DataFrame([], index=np.linspace(0.2, 0.6, 5))
    for i in ['small', 'middle', 'large']:
        for thresh in np.linspace(0.2, 0.6, 5):
            data = np.loadtxt(i + '_sample_' + str(thresh), delimiter=' ')
            for val in range(4):
                df.loc[thresh, i + str(val)] = np.count_nonzero(data == val)
    df.to_csv('analysis.csv')


def dist_matrix_1d(nbin=100):
    '''
    Generate distance matrix for earth mover's distance use
    '''

    #dist_matrix = np.abs(np.subtract.outer(range(nbin), range(nbin)))
    dist_matrix = pd.DataFrame()
    for x in range(nbin):
        for y in range(nbin):
            dist_matrix.loc[x, y] = abs(x-y)
    return dist_matrix.values


def nn_bulk_modulus_single(baseline_id, data, n_nn=1, emd='both'):
    '''
    Using the calucated EMD values and find the n nearest neighbor to estimate 
    bulk modulus

    Args:
        baseline_id: the mp task_id need to be estimated
        data: mp json data from file
        n_nn: number of nearest neighbors
        emd: the emd values used
            rdf: only rdf emd
            compos: only composition emd
            both:
    Return:
        The baseline mp-id
        Ground truth of bulk modluls
        mp-id of the nearest neighbor
        Estimation of the bulk modulus
        EMD distance between the nearest neighbor and baseline
    '''
    df_mp = pd.DataFrame.from_dict(data)
    df_mp = df_mp.set_index('task_id')

    if emd == 'rdf' or emd == 'both':
        rdf_emd_file = os.path.join(sys.path[0], '../rdf_emd/', baseline_id + '_emd.csv')
        rdf_emd = pd.read_csv(rdf_emd_file, index_col=0)

    if emd == 'compos' or emd == 'both':
        compos_emd_file = os.path.join(sys.path[0], '../compos_emd/', baseline_id + '_compos_emd.csv')
        compos_emd = pd.read_csv(compos_emd_file, index_col=0)

    # now add two emds to give a finally emd, need new method
    if emd == 'both':
        total_emd = rdf_emd.add(compos_emd)
    elif emd == 'rdf':
        total_emd = rdf_emd
    elif emd == 'compos':
        total_emd = compos_emd
        
    nn_mps = total_emd.nsmallest(n_nn + 1, baseline_id)

    # sometimes there are other structures have zero emd with 
    # the baseline_id, so it is not always the first row is the baseline_id
    # so use drop instead of removing the first row
    try:
        nn_mps = nn_mps.drop(baseline_id)
    except:
        # sometimes even the baseline_id may not be in the n-nearest
        nn_mps = nn_mps[1:]

    aver_modul = 0
    for task_id in nn_mps.index.values:
        aver_modul += df_mp['elasticity.K_VRH'][task_id]

    return (baseline_id, df_mp['elasticity.K_VRH'][baseline_id], 
            nn_mps.index.values[0], aver_modul / n_nn, 
            nn_mps.values[0][0] )


def nn_bulk_modulus_matrix_add(data, nn=1, simi_dir = '.', simi_matrix=['extended_rdf_emd'], scale=True):
    '''
    Deprecated. The adding similarity matrix is worse than currently used
    Using the calucated EMD values and find the n nearest neighbor to estimate 
    bulk modulus. If two EMD matrix are used, then add the values of the two matrix
    
    Args:
        data: mp json data from file
        nn: number of nearest neighbors
        matrix: a list of the emd values used
            extended_rdf: emd of extended rdf
            original_rdf:
            composition: only composition emd
        scale: no implement yet
    Return:
        The baseline mp-id
        Ground truth of bulk modluls
        Estimation of the bulk modulus
        EMD distance between the nearest neighbor and baseline
        mp-id of the nearest neighbor
    '''

    warnings.warn('`nn_bulk_modulus_matrix_add` is deprecated and will be removed', DeprecationWarning)

    # read the modulus data and convert to pandas dataframe
    # for the usage of later ground truth and mp ids
    df_mp = pd.DataFrame.from_dict(data)
    df_mp = df_mp.set_index('task_id')

    # read the emd matrix
    total_emd = pd.DataFrame(np.zeros((len(data), len(data))),
                            index=df_mp.index, columns=df_mp.index)
    if 'extended_rdf_emd' in simi_matrix:
        rdf_emd_file = os.path.join(simi_dir, 'extended_rdf_emd.csv')
        rdf_emd = pd.read_csv(rdf_emd_file, index_col=0)
        rdf_emd = rdf_emd.fillna(0).add(rdf_emd.transpose().fillna(0))
        if scale:
            total_emd = total_emd.add(rdf_emd * 10)
        else:
            total_emd = total_emd.add(rdf_emd)

    if 'extended_rdf_cosine' in simi_matrix:
        rdf_emd_file = os.path.join(simi_dir, 'extended_rdf_cosine.csv')
        rdf_emd = pd.read_csv(rdf_emd_file, index_col=0)
        rdf_emd = rdf_emd.fillna(0).add(rdf_emd.transpose().fillna(0))
        total_emd = total_emd.add(rdf_emd)

    if 'composition_emd' in simi_matrix:
        compos_emd_file = os.path.join(simi_dir, 'composition_emd.csv')
        compos_emd = pd.read_csv(compos_emd_file, index_col=0)
        compos_emd = compos_emd.fillna(0).add(compos_emd.transpose().fillna(0))
        total_emd = total_emd.add(compos_emd)

    if total_emd.isna().all().sum() > 0:
        raise IndexError('data file is a different size to distance matrix')

    pred_k = pd.DataFrame(index=df_mp.index,
                        columns=['ground_truth', 'predict_bulk_modulus', 
                                    'smallest_emd', 'nearest_neighbor'])
    for baseline_id in df_mp.index.values:
        nn_mps = total_emd[baseline_id].drop(baseline_id).nsmallest(nn)

        aver_modul = 0
        for task_id in nn_mps.index.values:
            aver_modul += df_mp['elasticity.K_VRH'][task_id]

        # df.loc index go first, df[] column go first
        pred_k.loc[baseline_id]['predict_bulk_modulus'] = aver_modul / nn
        pred_k.loc[baseline_id]['ground_truth'] = df_mp['elasticity.K_VRH'][baseline_id]
        pred_k.loc[baseline_id]['smallest_emd'] = nn_mps.values[0]
        pred_k.loc[baseline_id]['nearest_neighbor'] = nn_mps.index.values[0]
        

    return pred_k


def nn_bulk_modulus_matrix_step(data, simi_matrix=['extended_rdf_emd','composition_emd']):
    '''
    Using the calucated EMD values and find the n nearest neighbor to estimate 
    bulk modulus

    If two similaity matrix is used, first search the nearest neighbors of the first matrix, 
    if same values are found, then use the second matrix to find which is the nearest

    Args:
        data: mp json data from file
        emd: a list of the emd values used, the first one will be used for the
            first nearest neighbor search. values:
                extended_rdf: emd of extended rdfs
                extended_rdf_cos: cosine similarity of the extended rdfs
                composition: only composition emd
    Return:
        The baseline mp-id
        Ground truth of bulk modluls
        Estimation of the bulk modulus
        EMD distance between the nearest neighbor and baseline
        mp-id of the nearest neighbor
    '''
    # read the modulus data and convert to pandas dataframe
    # for the usage of later ground truth and mp ids
    df_mp = pd.DataFrame.from_dict(data)
    df_mp = df_mp.set_index('task_id')

    # read the emd matrix
    total_emd = pd.DataFrame(np.zeros((len(data), len(data))),
                            index=df_mp.index, columns=df_mp.index)
    
    file_1 = os.path.join(sys.path[0], simi_matrix[0]+'.csv')
    simi_mat_1 = pd.read_csv(file_1, index_col=0)
    simi_mat_1 = simi_mat_1.fillna(0).add(simi_mat_1.transpose().fillna(0))

    file_2 = os.path.join(sys.path[0], simi_matrix[1]+'.csv')
    simi_mat_2 = pd.read_csv(file_2, index_col=0)
    simi_mat_2 = simi_mat_2.fillna(0).add(simi_mat_2.transpose().fillna(0))

    pred_k = pd.DataFrame(index=df_mp.index,
                        columns=['ground_truth', 'predict_bulk_modulus', 
                                    'emd_1', 'emd_2', 'nearest_neighbor'])
    for baseline_id in df_mp.index.values:
        single_emd = simi_mat_1[baseline_id].drop(baseline_id)
        # fount the indice with the smallest value
        nn_idx = single_emd.where(single_emd == single_emd.min()).dropna()
        nn_idx = nn_idx.to_frame()
        
        if len(nn_idx) > 1:
            # if there are multiple structures have the same end,
            # then determine the nearest by considering the second similarity matrix
            new_list = []
            for task_id in nn_idx.index.values:
                new_list.append(simi_mat_2[baseline_id][task_id])
            nn_idx['new'] = new_list
            nn_mps = nn_idx.nsmallest(1, 'new')
        else:
            nn_idx['new'] = [0]
            nn_mps = nn_idx

        # df.loc index go first, df[] column go first
        pred_k.loc[baseline_id]['predict_bulk_modulus'] = df_mp['elasticity.K_VRH'][nn_mps.index.values[0]]
        pred_k.loc[baseline_id]['ground_truth'] = df_mp['elasticity.K_VRH'][baseline_id]
        pred_k.loc[baseline_id]['emd_1'] = nn_mps.values[0][0]
        pred_k.loc[baseline_id]['emd_2'] = nn_mps.values[0][1]
        pred_k.loc[baseline_id]['nearest_neighbor'] = nn_mps.index.values[0]

    return pred_k


def composition_similarity(baseline_id, data, index='z_number_78'):
    '''
    Calcalate the earth mover's distance between the baseline structure and all others
    in the dataset 

    Args:
        data: a pandas dataframe, index is mp-ids, and columns is element symbols
    Return:
        a pandas dataframe of pairwise distance between baseline id and all others
    '''
    # define the indice by call element_indice function
    element_indice()
    
    dist = []
    elem_similarity_file = os.path.join(sys.path[0], 'similarity_matrix.csv')
    dist_matrix = pd.read_csv(elem_similarity_file, index_col='ionA')
    dist_matrix = 1 / (np.log10(1 / dist_matrix + 1))

    if index == 'pettifor':
        dist_matrix = dist_matrix.reindex(columns=gridrdf.composition.pettifor, index=gridrdf.composition.pettifor) 
    # the earth mover's distance package need order C and float64 astype('float64')
    dist_matrix = dist_matrix.values.copy(order='C')

    compo_emd = pd.DataFrame(columns=[baseline_id])
    mp_id_1 = baseline_id
    for mp_id_2 in tqdm(data.index, mininterval=60):
        compo_emd.loc[mp_id_2] = emd(data.loc[mp_id_1].values.copy(order='C'), 
                                    data.loc[mp_id_2].values.copy(order='C'), 
                                    dist_matrix)
    return compo_emd


def composition_similarity_matrix(data, indice=None, index='z_number_78', elem_similarity_file='similarity_matrix.csv'):
    '''
    Calcalate pairwise earth mover's distance of all compositions, the composition should be 
    a 78-element vector, as the elemental similarity_matrix is 78x78 matrix in the order of  
    atom numbers

    Args:
        data: pandas dataframe of element vectors for all the structures
        index: see function element_indice for details
            z_number_78: (default) in the order of atomic number, this is default 
                because the similarity matrix is in this order
            z_number: see periodic_table in function element_indice
            pettifor: see function element_indice for details
            modified_pettifor: see element_indice
            elem_present: the vector only contain the elements presented in the dataset
    Return:
        a pandas dataframe of pairwise EMD with mp-ids as index
    '''
    # define the indice by call element_indice function
    element_indice()

    # if indice is None, then loop over the whole dataset
    if not indice:
        indice = [0, len(data)]
    
    dist = []
    dist_matrix = pd.read_csv(elem_similarity_file, index_col='ionA')
    dist_matrix = 1 / (np.log10(1 / dist_matrix + 1))

    if index == 'pettifor':
        dist_matrix = dist_matrix.reindex(columns=gridrdf.composition.pettifor, index=gridrdf.composition.pettifor) 
    # the earth mover's distance package need order C and float64 astype('float64')
    dist_matrix = dist_matrix.values.copy(order='C')
    
    ids = data.iloc[indice[0]:indice[1]].index
    compo_emd = pd.DataFrame(np.zeros((indice[-1], indice[-1])), index=ids, columns=ids )
    for i1 in range(indice[0], indice[1]):
        mp_id_1 = data.index[i1]
        for i2, mp_id_2 in enumerate(data.index):
            if i1 <= i2:
                emd_value = emd(data.loc[mp_id_1].values.copy(order='C'), 
                                data.loc[mp_id_2].values.copy(order='C'), 
                                dist_matrix)
                compo_emd.loc[mp_id_1, mp_id_2] = emd_value
            else:
                compo_emd.loc[mp_id_1, mp_id_2] = np.nan
    return compo_emd

@nb.njit(parallel=True)
def _emd_cumsum_row(cumsum_grid_array, index, bin_width, weights = None):
    """Super-fast implementation of EMD with Numba support for an array of structures.
    
    WARNING: This algorithm does not do any error checking! It assumes that the input is 
             a NumPy array of cumulative, normalised grid distributions.

    This is optimised to calculate EMD between one GRID and many others in parallel, with
    the aim to create a 2D EMD distance matrix. As such, it only computes EMD to structures
    with index greater or equal to `index`.


    Parameters
    ----------
    
    cumsum_grid_array: np.ndarray, shape (nstructures, ngrid_shells, nbins)
        Cumulative distributions of normalised GRIDs, stacked into a single 3D tensor. 
        IMPORTANT: cumulative distributions must all sum to the same value!

    index: int
        Index of structure to calculate pairwise EMDs for
    bin_width : float
        Width of each histogram bin, used to normalise the resulting EMD correctly
    weights: array-like, default None
        Weighting to apply to each grid shell before computing the mean. If None, equal weighting (1/n_shells) 
        will be applied.
        The weights should be the same length as the number of GRID shells, and should sum to the number of shells (not 1).


    Notes
    -----

    As detailed [scipy.stats.wasserstein_distance](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wasserstein_distance.html), 
    an equivalent definition of EMD based on cumulative distribution functions is:

    $$
    EMD(u,v) = \int_{-\infty}{+\infty}|U - V|
    $$
    where $u,v$ are probability distributions and $U, V$ are their respectives cumulative distributions. As such, the EMD between 
    two GRID representations $(G,H)$ will be
    $$
    EMD(G, H) = \sum_{j=1}^n \left( \frac{\sum_i | G_{j,i} - H_{j,i} |}{max(U)} \times w \right) / n
    $$
    where the sum over $i$ is for all bins in a given GRID shell (to the same cutoff) and $w$ is the width of each bin. The 
    division by $max(U)$ is needed in cases where $u$ and $v$ are not normalised to sum to 1, for instance when $w < 1$.
    $j$ is the sum across all GRID shells $n$, meaning that the overall EMD is the mean of each shell EMD (currently, this
    may change in future).

    """

    if weights is None:
        shell_weights = np.ones(cumsum_grid_array.shape[1])
    else:
        # Avoid numba type errors by simply multiplying ones by weights (also checks array sizes)
        shell_weights = np.ones(cumsum_grid_array.shape[1]) * weights

    output = np.zeros(cumsum_grid_array.shape[0])
    for j in nb.prange(index,cumsum_grid_array.shape[0]):
        output[j] = np.mean( shell_weights * np.sum(np.abs(cumsum_grid_array[index] - cumsum_grid_array[j]), axis=-1) / np.max(cumsum_grid_array[index]) * bin_width)
        #output[j] = np.sum(shell_weights)
    return output

@nb.njit()
def _EMD_cumsum(cumulative_grid1, cumulative_grid2, bin_width, weights=None):
    """
    Return EMD between two cumulative GRID (2D) representations.

    Parameters
    ----------

    cumulative_grid1, cumulative_grid2: array-like
        2D arrays containing cumulative distribution values with which to calculate Earth Mover's Distance
    bin_width: float
        Width of each bin in order to normalise the EMD correctly
    weights: array-like, default None
        Weighting to apply to each grid shell before computing the mean. If None, equal weighting (1/n_shells) 
        will be applied.
        The weights should be the same length as the number of GRID shells, and should sum to 1.

    Returns
    -------

    EMD : float
        Earth mover distance between two GRID distributions

    Notes
    -----

    For mathematical explanation see `_EMD_cumsum_row`

    """
    
    if weights is None:
        weights = np.ones(cumulative_grid1.shape[0])
    return np.mean( weights * np.sum(np.abs(cumulative_grid1 - cumulative_grid2), axis=-1) / np.max(cumulative_grid1) * bin_width)

@nb.njit()
def _flattened_EMD_cumsum(cumulative_grid_flat1, cumulative_grid_flat2, shape, bin_width, weights=None):
    """
    Compute EMD between two flattened GRID distributions.
    """

    grid1 = cumulative_grid_flat1.reshape(shape)
    grid2 = cumulative_grid_flat2.reshape(shape)

    return _EMD_cumsum(grid1, grid2, bin_width, weights)

def super_fast_EMD_matrix(grids,
                          bin_width,
                          weighting='constant',
                          weighting_kwargs={'n':0.0},
                          results_array = None,):
    """
    Efficient parallelised method to compute the matrix of EMD for a collection of GRID representations.

    Warning - for large lists of GRIDS, this can be very memory (and compute) intensive!

    Parameters
    ----------
    grids : array-like
        Iterable of GRID representations
    bin_width : float
        Width of each GRID bin for correctly normalising EMD
    weighting : str or function
        Weighting to apply to each grid shell when computing the overall EMD, allowing
        biasing towards short- or long-range similarity. A string argument will use 
        a pre-defined weighting function, or alternatively a callable can be provided in
        order to compute weightings. If callable, the function should take the number of 
        shells as input and return an array of values.
        Note that weights should be normalised to sum to the number
        of GRID shells (not 1).
        In-built options:
            'constant': equal weighting to all GRID shells
            'power' : weighting of shell x is 1/x**n (normalised over all shells)
        Default : 'constant' (equal weighting to bins)
    weighting_kwargs : dict
        Kwargs to use for computing weighting, either as parameters to in-built 
        functions or to be passed direct to a weighting callable.
    results_array : array-like, default None
        If not None, EMD values will be stored to this array. It should
        be square with shape (len(grids), len(grids)).
        Can be substituted with an HDF5 file object to minimise memory consumption.


    Notes
    -----

    - This method requires all GRIDs to be the same shape so that they can be converted into a 3-dimensional NumPy array.
    - Checks are made for normalisation
    """

    shapes = np.array([i.shape for i in grids])
    assert np.all(np.isclose(shapes, shapes[0])), "All GRIDs must have the same dimensions"

    if weighting == 'constant':
        # Use a constant weighting
        weights = np.ones(grids[0].shape[0])
    elif weighting == 'power':
        # Use x**(-n)
        assert 'n' in weighting_kwargs, "weighting_kwargs must contain `n` when weighting==power"
        power_series = np.power(np.arange(1, grids[0].shape[0]+1), -weighting_kwargs['n'])
        # Normalise the weights to n_grid_shells
        weights = power_series * grids[0].shape[0] / power_series.sum()
    elif callable(weighting):
        weights = weighting(grids[0].shape[0], **weighting_kwargs)


    if not isinstance(grids, np.ndarray):
        grid_input = np.array(grids)
    else:
        grid_input = grids

    # Perform cumulative summation for fast EMD calculation
    grid_cumsum = np.cumsum(grid_input, axis=-1)

    assert np.all(np.isclose(grid_cumsum[0,0,-1], grid_cumsum[:,:,-1])), "All GRIDs must be normalised to the same area. "

    if results_array:
        assert results_array.shape == (grid_input.shape[0], grid_input.shape[0])
        output = results_array
    else:
        output = np.zeros((grid_cumsum.shape[0], grid_cumsum.shape[0]))

    # Do the time-consuming calculation in parallel (with progress bar)
    for i in tqdm(range(grid_cumsum.shape[0]), mininterval=5):
        output[i] = _emd_cumsum_row(grid_cumsum, i, bin_width, weights=weights)
        # Fill in the lower diagonal (assumes we are iterating in order)
        output[i, :i] = output[:i, i]
    return output

def rdf_emd_similarity(rdf_a, rdf_b, max_distance=10, method='fast'):
    """
    Compute EMD similarity between two RDF distributions (1D or 2D)
    
    Notes
    -----
    
    This is the most time-consuming step in computing EMD similarity between distributions (particularly GRIDs).
    TODO: optimise this method further
    
    The "fast" method joins all GRID shells into a single 1D vector, and uses a modified distance metric so that
    comparisons between equivalent shells will give the same EMD, but distributions between shells have a high EMD contribution
    (**unless the distributions are not normalised**). This requires only one call to `wasserstein_distance` (with n_shell*max_dist bins)
    rather than n_shell calls to `wasserstein_distance` (each with max_dist bins).
    
    """
    
    assert rdf_a.shape == rdf_b.shape
    emd_bins = np.linspace(0, max_distance, rdf_a.shape[-1])
    
    if len(rdf_a.shape) == 2:
        if method=='orig':
            shell_distances = []
            for j in range(rdf_a.shape[0]):
                shell_distances.append(wasserstein_distance(emd_bins, emd_bins, 
                                                            rdf_a[j], rdf_b[j]))
                                                            

            return np.array(shell_distances).mean()
        elif method=='fast':
            assert np.isclose(rdf_a.sum(axis=1), 1.0).all(), "All GRID shells must be normalized to 1.0 in order for the fast method to work correctly."
            assert np.isclose(rdf_b.sum(axis=1), 1.0).all(), "All GRID shells must be normalized to 1.0 in order for the fast method to work correctly."
            shell_offsets = np.arange(0, rdf_a.shape[0]) * emd_bins[-1]
            emd_bins_1D = (np.vstack([emd_bins]*rdf_a.shape[0]) + np.column_stack([shell_offsets] * rdf_a.shape[-1]))
            if isinstance(rdf_a, np.ndarray) and isinstance(rdf_b, np.ndarray):
                # Using standard NumPy array processing
                emd_bins_1D = emd_bins_1D.ravel()
                return wasserstein_distance(emd_bins_1D, emd_bins_1D, rdf_a.ravel(), rdf_b.ravel())
            elif isinstance(rdf_a, coo_matrix) and isinstance(rdf_b, coo_matrix):
                # We are using sparse array, so can gain a significant speedup (~2-5-fold)
                emd_bins_a = emd_bins_1D[rdf_a.row, rdf_a.col]
                emd_bins_b = emd_bins_1D[rdf_b.row, rdf_b.col]
                return wasserstein_distance(emd_bins_a, emd_bins_b, rdf_a.data, rdf_b.data)
    elif len(rdf_a.shape) == 1:
        return wasserstein_distance(emd_bins, emd_bins, rdf_a, rdf_b)
                                                            
    
    
def rdf_row_similarity(baseline_rdf, all_rdf, max_distance = 10):
    '''
    Calculate the earth mover's distance between two RDFs
    Current support vanilla rdf and extend rdf
    Using Guassian smearing for rdf

    Args:
        all_rdf
        baseline_rdf: the rdf need to be estimated
    Return:
        a pandas dataframe of pairwise distance between baseline id and all others
        for multiple shells rdf the distance is the mean value of all shells
    '''
    # used for wasserstein distance
    emd_bins = np.linspace(0, 10, 101)

    rdf_1 = baseline_rdf
    rdf_emd = pd.DataFrame(columns=['baseline'])

    if len(rdf_1.shape) == 2:
        dist_matrix = dist_matrix_1d(len(rdf_1[0]))
        for d, rdf_2 in enumerate(all_rdf):
            #rdf_len = np.array([len(rdf_1), len(rdf_2)]).min()
            #shell_distances = []
            #for j in range(rdf_len):
            #    shell_distances.append(wasserstein_distance(emd_bins, emd_bins, rdf_1[j], rdf_2[j]))
            #    #shell_distances.append(emd(rdf_1[j], rdf_2[j], dist_matrix))
            #rdf_emd.loc[d] = np.array(shell_distances).mean()
            rdf_emd.loc[d] = rdf_emd_similarity(rdf_1, rdf_2, max_distance=max_distance)

    elif len(rdf_1.shape) == 1:
        dist_matrix = dist_matrix_1d(len(rdf_1))
        for d, rdf_2 in enumerate(all_rdf):
            rdf_emd.loc[d] = wasserstein_distance(emd_bins, emd_bins, rdf_1, rdf_2)
            #rdf_emd.loc[d['task_id']] = emd(rdf_1, rdf_2, dist_matrix)
    
    return rdf_emd


def rdf_similarity_matrix(all_rdf,
                          data=None,
                          indice=None,
                          method='emd',
                          max_distance = 10,
                          ):
    '''
    Calculate the earth mover's distance between all RDF pairs, returning a square matrix.

    Parameters
    ----------

    all_rdf : array-like
        Iterable of RDF representations to use for RDF calculation
    data: iterable of dicts, optional
        Iterable the same length as all_rdf, containing the key `task_id` to be used to
        label the resulting DataFrame.

    Returns
    -------

    dissimilarity : DataFrame
        DataFrame containing pairwise dissimilarity between RDF representations.

    Notes
    -----

    `rdf_similarity_matrix` has changed to use `super_fast_EMD_matrix` internally, which is better
    parallelised. More control over data storage (e.g. saving direct to a file to reduce memory) can
    be achieved by using `super_fast_EMD_matrix` directly.
    '''
    
    warnings.warn('`rdf_similarity` is now effectively parallelised using numba, so indice is no longer used.', SyntaxWarning)

    rdf_lengths = [i.shape[0] for i in all_rdf]
    assert len(set(rdf_lengths)) == 1, "All RDFs (or GRIDs) must have the same number of bins - use `data_prepare.trim_rdf_bins` first"
    
    # Compute bin_width based on max distance
    bin_width = max_distance / rdf_lengths[0]

    if len(all_rdf[0].shape) == 2:
        # typically for extend RDF
        
        if data:
            ids = [i['task_id'] for i in data]
        else:
            ids = np.arange(len(all_rdf))
        
        dissimilarity = super_fast_EMD_matrix(all_rdf,
                                              bin_width,
                                              weighting='constant',
                                              weighting_kwargs={'n':0.0},
                                              results_array = None,)
        
        df = pd.DataFrame(dissimilarity, index=ids, columns=ids)

        # df = pd.DataFrame(np.zeros((len(ids), len(ids))), index=ids, columns=ids)
        # for i1 in tqdm(range(indice[0], indice[1]), desc='similarity matrix rows', mininterval=10):
        #     d1 = data[i1]
        #     for i2, d2 in enumerate(data):
        #         if i1 <= i2:
        #             df.loc[d1['task_id'], d2['task_id']] = rdf_emd_similarity(all_rdf[i1], all_rdf[i2], max_distance=max_distance)
        #         else:
        #             df.loc[d1['task_id'], d2['task_id']] = np.nan
    elif len(all_rdf[0].shape) == 1:
        # typically for vanilla RDF and other 1D input
        df = pd.DataFrame([])
        for i1 in tqdm(range(indice[0], indice[1]), desc='similarity matrix rows', mininterval=10):
            d1 = data[i1]
            for i2, d2 in enumerate(data):
                if i1 <= i2:
                    if method == 'emd':
                        shell_distance = rdf_emd_similarity(all_rdf[i1], all_rdf[i2], max_distance=max_distance)
                        
                    elif method == 'cosine':
                        shell_distance = spatial.distance.cosine(all_rdf[i1], all_rdf[i2])

                    df.loc[d1['task_id'], d2['task_id']] = shell_distance
                else:
                    df.loc[d1['task_id'], d2['task_id']] = np.nan
    return df 


def rdf_similarity_matrix_old(data, all_rdf, order=None, method='emd'):
    '''
    Deprecated. Have some methods there no longer used
        1. (reciprical) inner product as a similarity measure
        2. emd method implemented in pyemd
        3. the matrix in the order of lattice parameter or symmetry

    Calculate the earth mover's distance between two RDFs
    Current support vanilla rdf and extend rdf
    Using Guassian smearing for rdf

    Args:
        data: data from json
        all_rdf: 
        order: how the compounds are arranged
            None: using the order in the josn 'data'
            symmetry: in the order of space group number, typically for 
                influence of distortion
            lattice: in the order of lattice constant
        method: how similarity is calculated
            emd: earth mover's distance
            linear: inner product
            linear-reciprocal: reciprocal of inner product (sum), useful for visualization
                and comparison with emd results
            cosine: cosine similarity implemented in scipy
            cosine-reciprocal: reciprocal of cosine
    Return:
        a pandas dataframe with all pairwise distance
        for multiple shells rdf the distance is the mean value of all shells
    '''

    warnings.warn("`rdf_similarity_matrix_old` is deprecated, and will be removed in a future release.", DeprecationWarning)

    # used for wasserstein distance
    emd_bins = np.linspace(0, 10, 101)

    if len(all_rdf[0].shape) == 2:
        # typically for extend RDF
        df = pd.DataFrame([])
        dist_matrix = dist_matrix_1d(len(all_rdf[0][0]))
        for i1, d1 in enumerate(tqdm(data, desc='rdf similarity', mininterval=60)):
            for i2, d2 in enumerate(data):
                if i1 <= i2:
                    rdf_len = np.array([len(all_rdf[i1]), len(all_rdf[i2])]).min()
                    shell_distances = []
                    for j in range(rdf_len):
                        if method == 'emd':
                            #shell_distances.append(wasserstein_distance(emd_bins, emd_bins, 
                            #                                            all_rdf[i1][j], all_rdf[i2][j]))
                            shell_distances.append(emd(all_rdf[i1][j], all_rdf[i2][j], dist_matrix))
                        elif method == 'linear' or method == 'linear-reciprocal':
                            shell_distances.append(np.inner(all_rdf[i1][j], all_rdf[i2][j]))
                        elif method == 'cosine' or method == 'cosine-reciprocal':
                            shell_distances.append(spatial.distance.cosine(all_rdf[i1][j], all_rdf[i2][j]))

                    if method == 'emd' or method == 'linear' or method == 'cosine':
                        df.loc[d1['task_id'], d2['task_id']] = np.array(shell_distances).mean()
                        df.loc[d2['task_id'], d1['task_id']] = np.array(shell_distances).mean()
                    elif method == 'linear-reciprocal' or method == 'cosine-reciprocal':
                        # add a small number 1e-11 to avoid dividing by zero
                        df.loc[d1['task_id'], d2['task_id']] = 1.0 / (np.array(shell_distances).mean() + 1e-11)
                        df.loc[d2['task_id'], d1['task_id']] = 1.0 / (np.array(shell_distances).mean() + 1e-11)    

    elif len(all_rdf[0].shape) == 1:
        # typically for vanilla RDF and other 1D input
        df = pd.DataFrame([])
        dist_matrix = dist_matrix_1d(len(all_rdf[0]))
        for i1, d1 in enumerate(tqdm(data, desc='rdf similarity', mininterval=60)):
            for i2, d2 in enumerate(data):
                if i1 <= i2:
                    if method == 'emd':
                        # see above for explanation and comparision of these two methods
                        shell_distance = wasserstein_distance(emd_bins, emd_bins, 
                                                            all_rdf[i1], all_rdf[i2])
                        # shell_distance = emd(all_rdf[i1], all_rdf[i2], dist_matrix)
                    elif method == 'linear' or method == 'cosine':
                        shell_distance = np.inner(all_rdf[i1], all_rdf[i2])
                    elif method == 'linear-reciprocal' or method == 'cosine-reciprocal':
                        # add a small number 1e-11 to avoid dividing by zero
                        shell_distance = 1.0 / (np.inner(all_rdf[i1], all_rdf[i2]) + 1e-11)
                    df.loc[d1['task_id'], d2['task_id']] = shell_distance
                    df.loc[d2['task_id'], d1['task_id']] = shell_distance    
    
    if order == 'symmetry':
        df2 = pd.DataFrame(columns=['sg'])
        for d in data:
            struct = Structure.from_str(d['cif'], fmt='cif')
            df2.loc[d['task_id']] = struct.get_space_group_info()[1]
        new_index = df2.sort_values('sg').index.values

        df = df.reindex(columns=new_index, index=new_index, fill_value=0)
        np.fill_diagonal(df.values, 0)

    elif order == 'lattice':
        df2 = pd.DataFrame(columns=['lattice'])
        for d in data:
            struct = Structure.from_str(d['cif'], fmt='cif')
            df2.loc[d['task_id']] = struct.lattice.a
        new_index = df2.sort_values('lattice').index.values

        df = df.reindex(columns=new_index, index=new_index, fill_value=0)
        np.fill_diagonal(df.values, 0)
    
    return df 


def rdf_similarity_visualize(data, all_rdf, mode, base_id=31):
    '''
    Visualization of rdf similarity results.
    Currently only used to investigate the influence of lattice constant

    Args:
        data: data from json
        all_rdf: 
        base_id: ID number for the baseline structure, i.e. all others structures are compared 
            to this structure
    Return:
        a pandas dataframe with all pairwise distance
        for multiple shells rdf the distance is the mean value of all shells
    '''

    df = pd.DataFrame([])
    if mode == 'similarity_different_shell':
        vis_rdf_shells = [0, 2, 6, 12, 14, 20, 26, 28]
        for i2, d2 in enumerate(data):
            for j in vis_rdf_shells:
                df.loc[d2['task_id'], str(j)] = wasserstein_distance(all_rdf[base_id][j], all_rdf[i2][j])
    
    elif mode == 'similarity_different_':
        dist_list = []
        mp_indice = []
        vis_ids = [0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60]
        for i2 in vis_ids:
            mp_indice.append(data[i2]['task_id'])

            rdf_len = np.array([len(all_rdf[base_id]), len(all_rdf[i2])]).min()
            shell_distances = []
            for j in range(rdf_len):
                shell_distances.append(wasserstein_distance(all_rdf[base_id][j], all_rdf[i2][j]))
            dist_list.append(shell_distances)

        df = pd.DataFrame(dist_list, index=mp_indice).transpose()
    
    elif mode == 'rdf_shell':
        rdf0s = []
        local_extrems = [21, 25, 29, 31, 36, 40]
        for i in local_extrems:
            rdf0s.append(all_rdf[i][0])
        df = pd.DataFrame(rdf0s, index=local_extrems).transpose()
    
    elif mode == 'rdf_shell_emd_path':
        emd_flows = []
        local_extrems = [21, 25, 29, 36, 40]
        dist_matrix = dist_matrix_1d(len(all_rdf[0][0]))

        for i in local_extrems:
            em = emd_with_flow(all_rdf[base_id][0], all_rdf[i][0], dist_matrix)
            print(em[0], wasserstein_distance(all_rdf[base_id][0], all_rdf[i][0]))
            emd_flows.append(em[1])
        df = pd.DataFrame(emd_flows, index=local_extrems)
    
    elif mode == 'lattice_difference':
        nums = [0, 259, 544]
        for base_id in nums:
            a1 = Structure.from_str(data[base_id]['cif'], fmt='cif').lattice.a
            for i2, d in enumerate(data):
                a2 = Structure.from_str(data[i2]['cif'], fmt='cif').lattice.a
                rdf_len = np.array([len(all_rdf[base_id]), len(all_rdf[i2])]).min()
                shell_distances = []
                for j in range(rdf_len):
                    shell_distances.append(wasserstein_distance(all_rdf[base_id][j], all_rdf[i2][j]))
                df.loc[data[i2]['task_id'], data[base_id]['task_id'] + 'lattice'] = a1 - a2
                df.loc[data[i2]['task_id'], data[base_id]['task_id']] = np.array(shell_distances).mean()
    
    return df


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Earth Mover Distance',
                                    formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--input_file', type=str, default='../MP_modulus.json',
                        help='the bulk modulus and structure from Materials Project')
    parser.add_argument('--output_file', type=str, default='',
                        help='currently for rdf similarity')
    parser.add_argument('--rdf_dir', type=str, default='./',
                        help='dir has all the rdf files')
    parser.add_argument('--simi_dir', type=str, default='./',
                        help='dir containing distance matrices')                        
    parser.add_argument('--task', type=str, default='rdf_similarity',
                        help='which property to be calculated: \n' +
                            '   rdf_similarity: \n' +
                            '   rdf_similarity_matrix: \n' +
                            '   composition_similarity_matrix: \n' +
                            '   composition_similarity: \n' +
                            '   rdf_similarity_visualize: \n' +
                            '   find_same_structure: find all structures giving same rdf \n' +
                            '   find_same_rdf: find all same rdf \n' +
                            '   nn_bulk_modulus: predict bulk modulus using nearest neighbor \n' +
                            '   '
                      )
    parser.add_argument('--baseline_id', type=str, default=None,
                        help='only used for single rdf_similarity composition_similarity tasks')
    parser.add_argument('--data_indice', type=str, default=None,
                        help='start and end indice of the sub dataset')
    parser.add_argument('-p', '--procs', type=int, default=None,
                        help='Number of processors to parallelize over where implemented')

    args = parser.parse_args()
    input_file = args.input_file
    output_file = args.output_file    
    rdf_dir = args.rdf_dir
    task = args.task
    baseline_id = args.baseline_id
    procs = args.procs
    
    if isinstance(args.data_indice, str):
        indice = list(map(int, args.data_indice.split('_')))
    else:
        indice = None

    with open (input_file,'r') as f:
        data = json.load(f)

    if task == 'rdf_similarity':
        if procs is not None and procs > 1:
            all_rdf = rdf_read_parallel(data, rdf_dir, procs=procs)
        else:
            all_rdf = rdf_read(data, rdf_dir)
        baseline_rdf = np.loadtxt(rdf_dir + '/' + baseline_id, delimiter=' ')
        rdf_emd = rdf_row_similarity(baseline_rdf, all_rdf)
        rdf_emd.to_csv(output_file + baseline_id + '_rdf_emd.csv')

    elif task == 'rdf_similarity_matrix':
        if procs is not None and procs > 1:
            all_rdf = rdf_read_parallel(data, rdf_dir, procs=procs)
        else:
            all_rdf = rdf_read(data, rdf_dir)

        df = rdf_similarity_matrix(all_rdf, data, indice=indice, method='emd')
        df.to_csv(output_file + '_whole_matrix.csv')

    elif task == 'composition_similarity':
        elem_vectors, elem_symbols = composition_one_hot(data)
        compo_emd = composition_similarity(baseline_id, elem_vectors)
        compo_emd.to_csv(output_file + baseline_id + '_compos_emd.csv')

    elif task == 'composition_similarity_matrix':
        elem_vectors, elem_symbols = composition_one_hot(data)
        compo_emd = composition_similarity_matrix(elem_vectors, indice=indice)
        if indice:
            compo_emd.to_csv(output_file + str(indice[0]) + '_' + str(indice[1]) + '.csv')
        else:
            compo_emd.to_csv(output_file + '_whole_matrix.csv')

    elif task == 'rdf_similarity_visualize':
        if procs is not None and procs > 1:
            all_rdf = rdf_read_parallel(data, rdf_dir, procs=procs)
        else:
            all_rdf = rdf_read(data, rdf_dir)

        for mode in ['rdf_shell_emd_path']:
            df = rdf_similarity_visualize(data, all_rdf, mode=mode)
            df.to_csv(output_file + '_' + mode + '.csv')

    elif task == 'find_same_structure':
        match_list = find_same_structure(data)
        with open(output_file, 'w') as f:
            json.dump(match_list, f, indent=1)

    elif task == 'find_same_rdf':
        if procs is not None and procs > 1:
            all_rdf = rdf_read_parallel(data, rdf_dir, procs=procs)
        else:
            all_rdf = rdf_read(data, rdf_dir)
        all_rdf = rdf_trim(all_rdf)
        X_data = rdf_flatten(all_rdf)
        match_list = find_same_rdf(all_rdf,data)
        with open(output_file, 'w') as f:
            json.dump(match_list, f, indent=1)

    elif task == 'nn_bulk_modulus':
        if baseline_id:
            # for single point calculation
            mp_ids = os.listdir(os.path.join(sys.path[0], '../rdf_emd/'))
            for f in mp_ids:
                print(nn_bulk_modulus_single(f.replace('_emd.csv', ''), data))
        else:
            pred_k = nn_bulk_modulus_matrix_add(data, simi_matrix=['extended_rdf_emd', 'composition_emd'])
            pred_k.to_csv(output_file + 'pred_k.csv')





import unittest
import os
import glob
import numpy as np
from sklearn.neighbors import KernelDensity
from scipy.sparse import coo_matrix
from scipy.stats import wasserstein_distance

from gridrdf import earth_mover_distance


FIXTURES_LOC = os.path.join(os.path.dirname(__file__), 'fixtures')

class TestEMD(unittest.TestCase):
    """ Test correct numerical operation of RDF EMD calculation"""
    def setUp(self):
        """
        Set up standard distributions for testing EMD calculations.

        Attributes
        ----------

        self.distances : dict
            Dictionary of arrays of distances
        self.rdf_1D : dict
            1D kernel density estimates for each of the distance sets
        self.rdf_2D : dict
            2D kernel density estimates (i.e. GRID) for each distance set
        self.rdf_2D_sparse : dict
            2D KDE for each distance set, stored as a scipy sparse array
        self.rdf_2D_cumsum : dict
            2D cumulative summation of distance GRIDS

        self.EMD_2D : dict of dicts
            Pairwise EMD values between different distance

        """
        # Set up some simple RDF bin edges
        self.max_dist=10
        self.bin_width = 0.1
        # Decimal place to use for accuracy
        self.decimal = 1
        self.bins = np.linspace(0,self.max_dist, int(self.max_dist/self.bin_width)+1)

        self.distances = []
        self.distances.append(np.array([[1.0,],
                                        [3.0,],
                                        [4.0,],
                                        [4.5,],
                                        [5.0,],
                                        [5.5,],
                                        [6.0,],
                                        [6.5,],
                                        [7.0,],
                                        [7.5,],
                                        ])
                             )
        self.distances.append(np.array([[1.5,],
                                        [3.5,],
                                        [4.5,],
                                        [5.0,],
                                        [5.5,],
                                        [6.0,],
                                        [6.5,],
                                        [7.0,],
                                        [7.5,],
                                        [8.0,],
                                        ])
                             )

        self.distances.append(np.array([[1.2, 1.8,1.8],
                                        [1.8, 2.1, 2.2],
                                        [1.95, 2.6, 2.6],
                                        [2.4, 2.9, 2.7],
                                        [2.6, 3.5, 3.4],
                                        [3.1, 3.8, 3.8],
                                        [2.1, 2.6, 2.7],
                                        [3.3, 3.9, 3.9],
                                        [3.7, 4.2, 4.2],
                                        [4.1, 4.5, 4.4]
                                        ])
                              )

        # A simple linear offset makes it easier to know what the correct EMD should be, 
        # as long as the bins do not overlap the edges of the range (which causes anomalies)
        self.distances.append(self.distances[-1] - 0.25)

        # Generate unbroadened 2D histograms
        self.rdf_2D_hist = []
        self.rdf_2D_hist_cumsum = []
        for distances in self.distances:
            dist_hist = [np.histogram(i, bins=self.bins, density=True)[0] for i in distances]
            self.rdf_2D_hist.append(np.array(dist_hist))
            self.rdf_2D_hist_cumsum.append(np.cumsum(self.rdf_2D_hist[-1], axis=-1))


        # Generate 1D KDE distributions manually
        self.rdf_1D = []
        for distances in self.distances:             
            KDE = KernelDensity(kernel='gaussian', bandwidth=0.1)
            log_dens = KDE.fit(distances.ravel()[:, np.newaxis]).score_samples(self.bins[:, np.newaxis])
            dens = np.exp(log_dens)
            self.rdf_1D.append(dens / dens.sum())

        # Generate 2D KDE dists
        self.rdf_2D_kde = []
        self.rdf_2D_sparse = []
        self.rdf_2D_cumsum = []
        for distance_set in self.distances:
            grid = []
            for shell in range(distance_set.shape[0]):
                log_dens = KDE.fit(distance_set[shell][:, np.newaxis]).score_samples(self.bins[:, np.newaxis])
                dens = np.exp(log_dens)
                dens = dens / dens.sum()
                grid.append(dens.copy())

            self.rdf_2D_kde.append(np.array(grid))

            self.rdf_2D_sparse.append(coo_matrix(self.rdf_2D_kde[-1]))

            self.rdf_2D_cumsum.append(np.cumsum(self.rdf_2D_kde[-1], axis=-1))

        # Set up known distances (mean of shells), and fill unknown with NaN
        self.EMD_2D = np.array([[0, 0.5, 2.125, 2.32833333333],
                                [0.5, 0, 2.545, 2.76166666666],
                                [2.125, 2.545, 0, 0.25],
                                [2.328333333333, 2.76166666666, 0.25, 0]])

    def test_pure_wasserstein(self):
        """ Check that wasserstein distance computed directly from distances is OK"""

        for i, dist1 in enumerate(self.distances):
            for j, dist2 in enumerate(self.distances):
                EMD_vals = []
                for shell in range(dist1.shape[0]):
                    EMD_vals.append(wasserstein_distance(dist1[shell], dist2[shell]))
                
                self.assertAlmostEqual(np.mean(EMD_vals), self.EMD_2D[i,j])


    def test_2D_pairwise_optimised_hist(self):
        """ Test optimised _emd_cumsum method with different 2D distributions. """

        for i, dist1 in enumerate(self.rdf_2D_hist_cumsum):
            for j, dist2 in enumerate(self.rdf_2D_hist_cumsum):
                if not np.isnan(self.EMD_2D[i,j]):
                    self.assertAlmostEqual(earth_mover_distance._EMD_cumsum(dist1, dist2, self.bin_width),
                                           self.EMD_2D[i,j],
                                           delta=self.bin_width)

    def test_2D_pairwise_hist(self):
        """ Check rdf_emd_similarity works for 2D case"""

        for i, dist1 in enumerate(self.rdf_2D_hist):
            for j, dist2 in enumerate(self.rdf_2D_hist):
                if not np.isnan(self.EMD_2D[i,j]):
                    # Check that the histograms are correct accounting for the bin_width
                    self.assertAlmostEqual(earth_mover_distance.rdf_emd_similarity(dist1, dist2, self.max_dist, method='orig'),
                                           self.EMD_2D[i,j],
                                           delta=self.bin_width)

                    
    def test_2D_pairwise_kde(self):
        """ Check rdf_emd_similarity works for 2D case"""

        for i, dist1 in enumerate(self.rdf_2D_kde):
            for j, dist2 in enumerate(self.rdf_2D_kde):
                if not np.isnan(self.EMD_2D[i,j]):
                    # Check that the EMDs are correct accounting for the bin_width
                    self.assertAlmostEqual(earth_mover_distance.rdf_emd_similarity(dist1, dist2, self.max_dist, method='orig'),
                                           self.EMD_2D[i,j],
                                           delta=self.bin_width)

    def test_2D_optimised_row(self):
        """ Check _emd_cumsum_row gives correct values. """

        EMDs = earth_mover_distance._emd_cumsum_row(np.array(self.rdf_2D_cumsum), 0, self.bin_width)
        # Test to the second decimal place
        np.testing.assert_array_almost_equal(EMDs, self.EMD_2D[0], decimal=self.decimal)

    def test_emd_duplicated(self):
        """ Test emd with the same RDF returns 0. """
        self.assertEqual(earth_mover_distance.rdf_emd_similarity(self.rdf_1D[0], self.rdf_1D[0]), 0)
        
    def test_emd_1D_simple(self):
        self.assertAlmostEqual(earth_mover_distance.rdf_emd_similarity(self.rdf_1D[0], self.rdf_1D[1], max_distance=self.max_dist), 0.5)

    def test_emd_1D_complex(self):
        self.assertAlmostEqual(earth_mover_distance.rdf_emd_similarity(self.rdf_1D[2], self.rdf_1D[3], max_distance=self.max_dist), 0.25)

    def test_emd_2D_method(self):
        """ Check original and faster 2D methods give same answer for complex case"""
        orig_method = earth_mover_distance.rdf_emd_similarity(self.rdf_2D_kde[2], self.rdf_2D_kde[3], max_distance=self.max_dist, method='orig')
        fast_method = earth_mover_distance.rdf_emd_similarity(self.rdf_2D_kde[2], self.rdf_2D_kde[3], max_distance=self.max_dist, method='fast')

        self.assertAlmostEqual(orig_method, fast_method)

    def test_emd_2D_sparse(self):
        """ Check fast and sparse 2D methods give same answer"""
        sparse_method = earth_mover_distance.rdf_emd_similarity(self.rdf_2D_sparse[2], self.rdf_2D_sparse[3], max_distance=self.max_dist, method='fast')
        fast_method = earth_mover_distance.rdf_emd_similarity(self.rdf_2D_kde[2], self.rdf_2D_kde[3], max_distance=self.max_dist, method='fast')

        self.assertAlmostEqual(sparse_method, fast_method)        

    def test_emd_matrix(self):
        test_dists = earth_mover_distance.super_fast_EMD_matrix(self.rdf_2D_hist,
                                                                self.bin_width,
                                                                weighting='constant',)
        
        np.testing.assert_almost_equal(test_dists, self.EMD_2D, decimal=self.decimal)

        
class TestHelperFunctions(unittest.TestCase):

    def test_dist_matrix(self):
        dim = 10
        dist_mat = earth_mover_distance.dist_matrix_1d(dim)
        self.assertEqual(dist_mat.shape, (dim,dim))

        test_mat = np.abs(np.meshgrid(range(dim), range(dim))[0] - np.meshgrid(range(dim), range(dim))[1])

        np.testing.assert_allclose(dist_mat, test_mat)    
        
if __name__ == '__main__':
    unittest.main()


""" Read previously generated RDF or GRID data.
"""

import numpy as np
import gzip, tarfile
from tqdm import tqdm
import os
import multiprocessing as mp


def rdf_read(data, rdf_dir, zip_file=False):
    '''
    Read the rdf files from the dir.  

    Args:
        data: modulus data from materials project
        rdf_dir: the dir has all (and only) the rdf files
        zip_file: if the RDFs are gzip file
    Return:
        a list of np array (rdfs) with different length, 
        first dimension is the number of
        samples, and the second dimension is the flatten 1D rdf
    '''
    all_rdf = []
    for d in tqdm(data, desc='rdf read', mininterval=5):
        rdf = _rdf_single_read(d, rdf_dir, zip_file)
        all_rdf.append(rdf)
    return all_rdf
    
def _rdf_single_read(data_row, rdf_dir, zip_file=False):
    """ Read single data file """
    rdf_file = os.path.normpath(os.path.join(rdf_dir, data_row['task_id']))
    if zip_file:
        with gzip.open(rdf_file + '.gz', 'r') as f:
            rdf = np.loadtxt(f, delimiter=' ')
    else:
        rdf = np.loadtxt(rdf_file, delimiter=' ')
    return rdf

def _rdf_single_read_star(args):
    """ Make multiprocessing work with multiple arguments without starmap"""
    return _rdf_single_read(*args)
    
def rdf_read_parallel(data, rdf_dir, zip_file=False, procs=None):
    """ 
    Read rdf files in parallel using multiprocessing
    
    See rdf_read for (serial) documentation and usage
    
    """
    
    if __name__ == "__main__":
    
        print("Reading data in parallel...")
        
        pool = mp.Pool(procs)
        args = [(i, rdf_dir, zip_file) for i in data]
        rdf_collect = list(tqdm(pool.imap(_rdf_single_read_star, args), total=len(args)))
        
        pool.close()
        pool.join()
        
    else:
        print("Reading data in serial...")
        rdf_collect = rdf_read(data, rdf_dir, zip_file)
        
    print("Reading done.")
        
    return rdf_collect
    
    
    
    
def shell_similarity_read(data, rdf_dir):
    '''
    Read the rdf files from the dir.  

    Args:
        data: modulus data from materials project
        rdf_dir: the dir has all (and only) the rdf files
    Return:

    '''
    all_shell_simi = []
    for d in data:
        shell_simi_file = rdf_dir + '/../shell_similarity/' + d['task_id']
        shell_simi = np.loadtxt(shell_simi_file, delimiter=' ')
        all_shell_simi.append(shell_simi.flatten())
    return np.stack(all_shell_simi)


def rdf_read_tar(data, x_file):
    '''
    Read the rdf files from a tar file.  

    Args:
        data: modulus data from materials project
        x_file: the tar file has rdfs
    Return:
        a list of np array (rdfs) with different length, 
        first dimension is the number of
        samples, and the second dimension is the flatten 1D rdf
    '''
    all_rdf = []
    with tarfile.open(x_file, 'r:*') as tar:
        for d in data:
            rdf_file = d['task_id']
            rdf = np.loadtxt(tar.extractfile(rdf_file), delimiter=' ')
            all_rdf.append(rdf)
    return all_rdf


def read_all_fs():
    '''
    '''
    import os
    import numpy as np
    import pandas as pd

    fs = []
    for f in os.listdir('.'):
        fs.append(np.loadtxt(f, delimiter=' '))

    df = pd.DataFrame(fs).transpose()





""" Miscellaneous utility functions.
"""

def int_or_str(value):
    '''
    For augment parser trim, where input can be either integer or string
    '''
    try:
        return int(value)
    except:
        return value


def read_and_merge_similarity_matrix():
    import pandas as pd

    t = pd.DataFrame([])
    for i in range(0, 12000, 100):
        a = pd.read_csv(str(i)+'_'+str(i+100)+'.csv', index_col=0)
        t = pd.concat([t,a])

    a = pd.read_csv(str(12000)+'_'+str(12177)+'.csv', index_col=0)
    t = pd.concat([t,a])

    t.to_csv('total_matrix.csv')


"""
A script to re-create the bulk modulus results of the original GRID paper.

WARNING: Some aspects of this script may take some time.

The slowest aspect is calculating EMD dissimilarity between all pairs. Currently, computing the full EMD matrix (~12,000**2 pairs)
takes around 30 minutes on an intel i5-8500 (3.00 GHz, 6-cores) processor, using efficient parallelisation.

"""

__author__ = "James Cumby"
__email__ = "james.cumby@ed.ac.uk"


import gridrdf
import os
import json
import numpy as np
import sklearn

script_location = os.path.dirname(__file__)

results_location = os.path.join(script_location, 'testing')
if not os.path.exists(results_location):
    os.makedirs(results_location)

data_source_loc = os.path.join(script_location, '../data_sources/')
    

# File to use for initial structure data
# If MP_API_KEY variable is not set, this script will try to read
# data in from the file specified, otherwise it will write out the 
# materials extracted from the Materials Project.

# This is the file containing all bulk-moduli containing structures
#data_file_basename = 'MP_modulus.json'    

# This is a smaller 'test set' to use as an example.
data_file_basename = 'small_test_set.json'

# Step 1 - Import bulk moduli data from materials project
APIkey = os.environ.get('MP_API_KEY')
if APIkey is None:
    print("You need to provide your (legacy) Materials Project API key as the environment variable `MP_API_KEY` to use the most recent data from Materials Project.")
    print(f"\nReverting to use the included file `data_sources/{data_file_basename}`")
    data_file = os.path.join(data_source_loc, data_file_basename)
    
    with open(data_file, 'r') as f:
        data = json.load(f)
else:
    data_file = os.path.join(results_location, data_file_basename)
    print(f"Extracting bulk modulus data from Materials Project and saving to {data_file}" )
    
    data = gridrdf.data_prepare.get_MP_bulk_modulus_data(APIkey)
    with open(data_file, 'w') as f:
        json.dump(data, f)
    
print(f"Obtained {len(data)} entries with bulk moduli.")


GRID_location = os.path.join(results_location, 'GRIDS')

# Step 2 - Prepare GRIDs

if os.path.exists(GRID_location):
    inp = input(f'Folder {GRID_location} already exists - do you want to overwrite its contents (y/N)? ')
else:
    inp = 'y'

print("Calculating GRID representations...")    
if len(inp) > 0 and inp.lower()[0] == 'y':
    gridrdf.data_prepare.batch_rdf(data,
                                   max_dist=10,
                                   bin_width = 0.1,
                                   broadening=0.1,
                                   output_dir = GRID_location,
                                   normalize=True
                                   )
    print("Calculating Grids - Done.")                               
                         

# Step 3 - Remove any structures with fewer than 100 GRID shells
# Step 4 - remove negative bulk moduli
# Step 5 - filter elements > Bi
# These steps can be run separately (see README) but are combined here using `data_prepare.main`

# Need to set up periodic table definitions first
gridrdf.composition.element_indice()

data_subset = gridrdf.data_prepare.main(data_source = data_file,
                                          tasks = ['subset_composition', 'subset_property', 'subset_grid_len', ],
                                          output_dir = GRID_location,
                                          MP_API_KEY=APIkey,
                                          output_file = 'subset.json',
                                          num_grid_shells = 100,
                                          composition = {'elem': gridrdf.composition.periodic_table_78, 'type':'consist'},
                                          data_property = ('elasticity.K_VRH', 0, np.inf),
                                          #GRIDS = all_GRID,
                                         )
       

# update GRIDS to match subset by re-reading
print("Re-reading subset of GRIDs")
all_GRID = gridrdf.data_io.rdf_read(data_subset, rdf_dir = GRID_location)

# Step 6 - Calculate EMD between all pairs of structures
#### WARNING - THIS MIGHT BE TIME CONSUMING! ####
print(f"Calculating pairwise EMD similarity for {len(data_subset)} materials")
print("This may take some time")
similarity = gridrdf.earth_mover_distance.rdf_similarity_matrix(all_GRID, data_subset, method='emd')
similarity.to_csv(os.path.join(results_location, 'GRID_similarity_matrix.csv'))


# Step 7 - Calculate composition similarity
print("Calculating composition similarity")

# First, convert composition to vector encoding
elem_vectors, elem_symbols = gridrdf.composition.composition_one_hot(data_subset, method='percentage')

# Now computed EMD similarity based on "distances" between species contained in `similarity_matrix.csv`)
# This is essentially Pettifor distance, but with non-integer steps defined by data-mining of probabilities.
comp_similarity = gridrdf.earth_mover_distance.composition_similarity_matrix(elem_vectors, 
                                                                     elem_similarity_file = os.path.join(data_source_loc, 'similarity_matrix.csv')
                                                                     )
comp_similarity.to_csv(os.path.join(results_location, 'composition_similarity_matrix.csv'))
                                                                     

total_similarity = 10*similarity + comp_similarity
                                                                     
# Only the upper triangle of the matrix is computed, so we need to 
# make it symmetric and convert to NumPy array.
total_similarity = total_similarity.add(total_similarity.T, fill_value=0)
total_similarity.to_csv(os.path.join(results_location, 'total_similarity_matrix.csv'))

total_similarity = total_similarity.to_numpy()

# Step 7 - Use KNN-regressor model to calculate nearest neighbours
# This produces the calculated vs predicted plot reported in the paper.
print("Calculating truth vs prediction data for bulk modulus with K-neighbours regression...")
K_data = np.array([ x['elasticity.K_VRH'] for x in data_subset ])
model = sklearn.neighbors.KNeighborsRegressor(n_neighbors=1, metric='precomputed')
gridrdf.train.calc_obs_vs_pred_2D(funct = model,
                                 X_data = total_similarity,
                                 y_data = K_data,
                                 test_size = 0.2,
                                 outdir= results_location,
                                )
                                
print('Done.')
                                
# Step 8 - Calculate learning curve
print('Calculating learning curve with K-neighbours regression')
gridrdf.train.calc_learning_curve(funct = model,
                                  X_data = total_similarity, 
                                  y_data = K_data,
                                  test_size=0.1,
                                  procs=1,
                                  output_dir = results_location,
                                  )
   
print('Done')
   
 